/***************************************************************************
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "canonical_string.h"
#include "chew.h"
#include "io.h"
#include "symbol.h"

using namespace std;

/*! \file canonical_string.cpp
	This file implements:-
	- class \e canonical<symbol>
	- class \e canonical<argument_list>
	- class \e canonical<std::string>
*/

/*! Explicitly construct a \c canonical<symbol> from a string
    identifier.

    \param  start  The start of the identifier.
    \param  pend    Either NULL or a pointer at which shall be
        stored on return the first address address beyond the
        parsed identifier.
*/
template<>
canonical<symbol>::canonical(char const *start, char const **pend)
{
	if (!start) {
		return;
	}
	if (*start && symbol::is_start_char(*start)) {
		_canonical.append(1,*start);
		start = chewer<continuation>(++start);
		for ( ; *start && symbol::is_valid_char(*start);
		      start = chewer<continuation>(++start)) {
			_canonical.append(1,*start);
		}
	}
	if (pend) {
		*pend = start;
	}
}

template<>
canonical<argument_list>::canonical(char const *start, char const **pend)
{
	if (!start) {
		return;
	}
	for (	; argument_list::is_valid_char(*start);
	        start = chewer<continuation>(++start)) {
		_canonical.append(1,*start);
	}
	if (pend) {
		*pend = start;
	}
}

void canonical<std::string>::parse_part(char const **start, size_t *len)
{
	char const *cp = *start;
	char const *advance = chewer<continuation>(cp);
	if (advance != cp) {
		*start = advance;
		*len = 0;
	} else {
		size_t read_off = io::read_offset(cp);
		advance = chewer<discardable>(cp);
		cp = io::read_pos(read_off);
		if (advance != cp) {
			*start = advance;
			*len = advance - cp;
		} else {
			read_off = io::read_offset(cp);
			advance = chewer<unbroken_string>(cp);
			cp = io::read_pos(read_off);
			*len = advance - cp;
			*start = cp;
		}
	}
}

canonical<string>::canonical(char const *start, char const **pend)
{
	if (!start) {
		return;
	}
	char const *cp = start;
	chewer_base::set_stop_at_quote(true);
	cp = chewer<discardable>(cp);
	for (	;;	) {
		char const * save_cp = cp;
		size_t tok_len;
		parse_part(&cp,&tok_len);
		if (save_cp == cp) {
			if (!tok_len) {
				break;
			}
			for (cp += tok_len;	save_cp < cp; ++save_cp) {
				if (isprint(*save_cp)) {
					_canonical += *save_cp;
				}
			}
		} else if (tok_len > 0) {
			_canonical.append(size_t(1),' ');
		}
	}
	if (pend) {
		*pend = cp;
	}
	// Trim trailing space
	size_t len = _canonical.length();
	if (len && _canonical[len - 1] == ' ') {
		for ( --len; len && _canonical[len] == ' '; --len) {}
		_canonical.resize(len + 1);
	}
	chewer_base::set_stop_at_quote(false);
}

/* EOF */
