#ifndef CANONICAL_STRING_H
#define CANONICAL_STRING_H
/***************************************************************************
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "platform.h"
#include "parameter_list.h"
#include <string>

struct symbol;

/*! \file canonical_string.h
	This file defines:-
	- class \c canonical_base
	- class template \c canonical<what>
	- class canonical<std::string>
*/

//! The base for classes representing canonical forms of types.
struct canonical_base {
	//! Cast the canonical representation to string.
	operator std::string () const {
		return _canonical;
	}

#if CPP11_HAVE_DECL_DEFAULT && !CPP11_HAVE_NON_PUBLIC_DEFAULT_DECL
	//! Default constructor
	canonical_base() = default;
	//! Copy constructor
	canonical_base(canonical_base const &) = default;
	//! Assignment
	canonical_base & operator=(canonical_base const &) = default;
#endif

protected:

#if CPP11_HAVE_DECL_DEFAULT && CPP11_HAVE_NON_PUBLIC_DEFAULT_DECL
	//! Default constructor
	canonical_base() = default;
	//! Copy constructor
	canonical_base(canonical_base const &) = default;
	//! Assignment
	canonical_base & operator=(canonical_base const &) = default;
#endif
	//! Destructor
	~canonical_base() {};

	//! String storing the canonical representation of a value.
	std::string _canonical;
};

/*! \brief Class template \e canonical<What> encapsulates the canonical
    representation of values of type \e What.

    \tparam What    The type that is canonically represented.
*/
template<class What>
struct canonical : canonical_base {
#if CPP11_HAVE_DECL_DEFAULT
	//! Default constructor
	canonical() = default;
#else
	canonical() {};
#endif

	/*! \brief Construct from a text pointer.

	    Explicitly construct a canonical representation given the
	    the start point and optional end point of some textual representation
		of a value.

	    \param  start  The start of the textual representation of the value.
	    \param  pend   Either NULL or a pointer at which shall be
	        stored on return the first address address beyond the textual
	        representation of the value.
	*/
	explicit canonical(char const *start, char const **pend = NULL);
};

/*! \brief Class \e canonical<std::string> encapsulatues the canonical
    representation of a string value parsed from a directive.

    The canonical representation is stripped of comments, line-continuations and
    linefeeds and all whitespace sequences are reduced to a single space.

*/
template<>
struct canonical<std::string> : canonical_base {
#if CPP11_HAVE_DECL_DEFAULT
	//! Default constructor
	canonical() = default;
#else
	canonical() {};
#endif
	/*! \brief Construct from a text pointer.

	    Explicitly construct a canonical representation of a string value
	    parsed from a directive.

	    \param  cp  The start of the string value.
	    \param  pend    Either NULL or a pointer at which shall be
	        stored on return the first address address beyond the string.
	*/
	explicit canonical(char const *cp, char const **pend = NULL);

private:

	/*! \brief Parse the next portion of a canonical string.

	    \param	start	On entry, points to the address of the
	    text to to be parsed. If on return the value stored
	    here is different then it is the address at which
	    parsing has stopped after consuming characters to be
	    discarded from the canonical form of the text.
	    If the stored value is unchanged then it is the
	    start of some text that belongs to the canonical form.

	    \param	len	The value stored here on entry is insignificant.
	    The meaning of the value stored here on return depends upon
	    the value stored at \e start:-

	    - If the value stored at \e start on return if different from
	    its entry value then discardable text has been parsed. In this
	    case, if the value at \e len is > 0 it is the length of discardable
	    whitespace and comments. If the value at \e len is 0
	    then a discardable sequence line-continuations has been parsed.

	    - If the value stored at \e start on return is unchanged from
	    its entry value then it may address text to be retained in the
	    canonical form. In this case, if the value stored at \e len
	    is > 0 it is the length of the text to be retained. If
	    the value stored at \e len is 0 then the string at \e start was
	    empty.

	    The function can be called with \e start initialised to
	    the address of a string to be parsed into canonical form.
	    On return, \e start will contain the address to be
	    passed to the next call to continue the parsing,
	    and the appropriate handling of the returned values can
	    be deduced from a comparsion of the entry and return
	    values at \e start, together with the value returned at
	    \e len:-

	    - When the value returned at \e start is changed and
	    the value stored at \e len is 0 then line-continuations
	    have been consumed and nothing need be appended to the
	    canonical form of the string.

	    - When the value returned at \e start is changed and
	    the value stored at \e len is > 0 then whitespace and/or
	    comments have been consumed and a single space should be
	    appended to the canonical form of the string.

	    - When the value returned at \e start is unchanged and
	    the value stored at \e len is 0 then parsing is done.

	    - When the value returned at \e start is unchanged and
	    the value stored at \e len is > 0 then then that number
	    of characters should be appended to the canonical form
	    from the address at \e start.

	*/
	static void
	parse_part(char const **start, size_t *len);
};

#endif /* EOF */
