#ifndef CHEWER_H
#define CHEWER_H

/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2002, 2003 Tony Finch <dot@dotat.at>.                   *
 *   All rights reserved.                                                  *
 *   Copyright (C) 1985, 1993 The Regents of the University of California. *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed by Mike Kinghan, imk@burroingroingjoing.com,              *
 *   derived from the code of Tony Finch                                   *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "platform.h"
#include <cstddef>

/*! \file chew.h
    This defines class \c chew
*/

//! A tag class for parameterizing class template \c chewer<What>
struct discardable {};
//! A tag class for parameterizing class template \c chewer<What>
struct continuation {};
//! A tag class for parameterizing class template \c chewer<What>
struct unbroken_string {};

/*! \brief Base class for instantiations of class template \c chewer<What>
    defining their common interface.

    The class is a base for types that perform scanning of text for
    a terminal condition in several settings.
*/
struct chewer_base {
	//! symbolic constants denoting comment states
	enum comment_state {
	    //! Outside a comment
	    NO_COMMENT = false,
	    //! In a C-style comment
	    C_COMMENT,
	    //! In a C++-style comment, i.e. between // and end of line
	    CXX_COMMENT,
	    //! Just after a forward-slash + back-slash + newline
	    STARTING_COMMENT,
	    //! Just after star + back-slash + newline in a C comment
	    FINISHING_COMMENT,
	    /*! In the text of a \c #error directive or the definiens of a
			\c \#define */
	    PSEUDO_COMMENT
	};

	//! Symbolic constants denoting states of a source line
	enum line_state {
	    //! Only whitespace and comments thus far on the line
	    LS_NEUTER,
	    //! Only whitespace, comments, and a hash thus far on the line
	    LS_DIRECTIVE,
	    /*! \brief Source code found on the line

	        Source code found on the line i.e text other than whitespace,
	        comments, or a directive */
	    LS_CODE
	};

	/*! \brief Cast the \c chewer object to \c const \c char \c *.

	    The operation consumes text starting at \c cp
	    and returns the first address not consumed.
	    returning the first address not consumed.
	*/
	operator char const * () const {
		return _end;
	}

	//! Initialize the class globals
	static void init() {
		_comment_state_ = NO_COMMENT;
		_in_double_quote_ = false;
		_line_state_ = LS_NEUTER;
		_last_quote_start_line_ = 0;
		_last_comment_start_line_ = 0;
		_stop_at_quote_ = false;
		_must_balance_quotes_ = true;
	}

	//! Say whether double-quotation is open in the chewed text
	static bool in_double_quote() {
		return _in_double_quote_;
	}

	//! Register the opening of double-quotation in the chewed text
	static void set_in_double_quote(bool in_quote) {
		_in_double_quote_ = in_quote;
	}

	//! Get the operative \c comment_state of the chewed text
	static comment_state get_comment_state() {
		return _comment_state_;
	}
	
	//!Set the operative \c comment_state of the chewed text
	static void set_comment_state(comment_state cs) {
		_comment_state_ = cs;
	}

	//! Get the operative \c line_state of the chewed text
	static line_state get_line_state() {
		return _line_state_;
	}

	//! Set the operative \c line_state of the chewed text
	static void set_line_state(line_state ls) {
		_line_state_ = ls;
	}

	/*! Get the line number on which quotation was last opened in the
	    chewed text.

	    \return The line number on which quotation was last opened in the
	    chewed text.
	*/
	static unsigned last_quote_start_line() {
		return _last_quote_start_line_;
	}

	//! Get the line number on which a comment last commenced in the chewed text
	static unsigned last_comment_start_line() {
		return _last_comment_start_line_;
	}

	//! Set the line number on which a comment last commenced in the chewed text
	static void set_last_comment_start_line();

	//! Say whether chewing stops at the opening of double-quotation.
	static bool stopping_at_quote() {
		return _stop_at_quote_;
	}

	//! Specify that chewing stops at the opening of double-quotation.
	static void set_stop_at_quote(bool stopping) {
		_stop_at_quote_ = stopping;
	}

	//! Say whether unbalanced double quotes are to be diagnosed
	static bool balancing_quotes() {
		return _must_balance_quotes_;
	}

	/*! \brief Specify that unbalanced double quotes are to be diagnosed,
	    or not.

	    \param  balancing   If true then unbalanced quotes are to be diagnosed,
			else not.
	*/
	static void set_balancing_quotes(bool balancing = true) {
		_must_balance_quotes_ = balancing;
	}

	/*! \brief Test for end of line.

	     Test for a newline sequence of either Unix type or Windows type at the
			text pointer.

	     \param	cp	Text pointer

	     \return 0 if \c cp does not address a newline sequence,
	     otherwise the length of the newline sequence, i,e,
	     1 for Unix '\n', 2 for Windows "\r\n"
	 */
	static unsigned eol(char const *cp) {
		if (cp[0] == '\n') {
			return 1;
		}
		if (cp[0] == '\r' && cp[1] == '\n') {
			return 2;
		}
		return 0;
	}
	
#if CPP11_HAVE_DECL_DEFAULT && !CPP11_HAVE_NON_PUBLIC_DEFAULT_DECL 
	//! Copy constructor
	chewer_base(chewer_base const &) = default;
	//! Assignment
	chewer_base & operator=(chewer_base const &) = default;
	//! Destructor
	~chewer_base() = default;
#endif

protected:

	//! Default constructor
	chewer_base()
		: _end(NULL) {};
#if CPP11_HAVE_DECL_DEFAULT && CPP11_HAVE_NON_PUBLIC_DEFAULT_DECL 
	//! Copy constructor
	chewer_base(chewer_base const &) = default;
	//! Assignment
	chewer_base & operator=(chewer_base const &) = default;
	//! Destructor
	~chewer_base() = default;
#endif


	/*! \brief Say whether a text pointer addresses a line-continuation.
	    \param  cp The text pointer.
	    \return True iff \c cp addresses a line-continuation.
	*/
	static bool is_line_continuation(char const *cp) {
		return *cp == '\\' && eol(cp + 1);
	}

	//! The current comment state
	static comment_state _comment_state_;
	//! Are we within double quotes?
	static bool _in_double_quote_;
	//! The current line state
	static line_state _line_state_;
	//! Line number of the most recent open-quote
	static unsigned	_last_quote_start_line_;
	//! Line number of most recent open-comment
	static unsigned	_last_comment_start_line_;
	//! Do we stop at quotes?
	static bool _stop_at_quote_;
	//! Do we have to balance quotation marks?
	static bool _must_balance_quotes_;

	//! The first address not consumed by the object.
	char const * _end;

};

/*! \brief Template class \c chewer<What> is a template text scanner that
    consumes controlled portions of input text in support of parsing.

    An instantiating class is responsible for maintaining parse-state
    with respect to comments and quotation.

    \tparam What The tag type that characterizes a \c chewer class

*/
template<class What>
struct chewer : chewer_base {
	//! Default constructor
	chewer() {};

	/*! \brief Construct from a text pointer.

	    Explicitly construct a \c chewer given a text pointer.
	    \param  cp  The text pointer at which to initialize the \c chewer
	*/
	explicit chewer(char const *cp);
};

/*! \brief Class \c chewer<discardable> is a text scanner
    that consumes contiguous whitespace and comments.
*/
template<>
struct chewer<discardable> : chewer_base {
	//! Default constructor
	chewer()
		:	_single_quote_length(0),_escape(false) {}

	/*! \brief Construct from a text pointer.

	    Explicitly construct a \c chewer given a text pointer.
	    \param  cp  The text pointer at which to initialize the \c chewer.

	    The constructor consumes contiguous whitespace and comments
	    starting at \c cp.
	*/
	explicit chewer(char const *cp)
		:	_single_quote_length(0),_escape(false) {
		_end = operator()(cp);
	}

	/*! \brief Cast the object to \c char \c const \c *.

	    \param  cp The address from which to scan text.
	    \return The The first address not scanned.

	    The operation consumes contiguous whitespace and comments
	    starting at \c cp and returns the first address not consumed.
	*/
	char const * operator()(char const *cp);

private:

	/*! \brief Current number of input characters commencing with the last '\''.

	    Current number of input characters commencing with the last '\'' that
	    possibly opens a quotation in single quotes, or 0 whenever we are
	    not within such a sequence. If this length ever exceeds 4 within we
	    conclude that the sequence is not a quotation but simply characters
	    following a free-standing independant apostrophe, and reset the value
	    to 0.

	    The sole reason for tracking this length  is to identify the 3-character
	    sequence '"' as a quotation of the double-quote character and avoid
	    falsely detecting the start or end of a quotation in double-quotes.
	    The equivalent 4-character sequence '\"' is unproblematic
	    because the double-quote will be seen to be escaped.
	 */
	size_t _single_quote_length;

	//! Last char read was escape?
	bool _escape;
};

#endif /* EOF*/
