/***************************************************************************
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "platform.h"
#include "diagnostic.h"
#include "chew.h"
#include "dataset.h"
#include "if_control.h"
#include "options.h"
#include "io.h"
#include <iostream>
#include <iomanip>
#include <cassert>
#include <cstdlib>

/*! \file diagnostic.cpp
    This file implements class \c diagnostic
*/

using namespace std;

unsigned diagnostic_base::_infos_(0);
unsigned diagnostic_base::_warnings_(0);
unsigned diagnostic_base::_errors_(0);
unsigned diagnostic_base::_abends_(0);
unsigned diagnostic_base::_error_directives_generated_(0);
unsigned diagnostic_base::_error_directives_operative_(0);

bool diagnostic_base::_in_code_(false);

list<diagnostic_base::ptr> diagnostic_base::_queue_;

diagnostic_base::diagnostic_base(std::string severity_name,
                                 severity::level level, unsigned id,
                                 bool in_code_only)
	: _gagged(false)
{
	unsigned reason = code(level,id);

	if (!_in_code_ && in_code_only) {
		_gagged = true;
		/* This message applies only when parsing code */
		return;
	}
	if (options::diagnostic_gagged(reason)) {
		/* This message is filtered out. Don't output*/
		_gagged = true;
		return;
	}
	_text << options::prog_name() << ": ";
	if (!options::got_opts()) {
		_text << "while parsing options: ";
	}
	if (!io::in_file_name().empty()) {
		_text << io::in_file_name() << ": ";
	}
	if (io::line_num()) {
		_text << "line " << io::line_num() << ": ";
	}
	_text << severity_name << ' ' << "0x" << setfill('0')
	      << hex << setw(5) << reason << ": " << dec;
}

void diagnostic_base::despatch() const throw(unsigned)
{
	count();
	if (!text().empty()) {
		cerr << text() << '\n';
	}
	if (level() == severity::abend) {
		exit(exitcode());
	}
	if (level() == severity::error) {
		throw(unsigned(*this));
	}
}

void diagnostic_base::epilogue()
{
	size_t infiles = dataset::files();
	char const * diagnostic_status = "";

	int ret = exitcode();
	if_control::top();
	io::top();

	if (_abends_) {
		diagnostic_status = " ABNORMALLY";
	} else if (_errors_) {
		diagnostic_status = " with errors";
	} else if (_warnings_) {
		diagnostic_status = " with warnings";
	} else if (_infos_) {
		diagnostic_status = " with remarks";
	}

	progress_summary_all_done() << "Completed"
		<< diagnostic_status << ", exit code 0x" << setfill('0') << setw(2) <<
		hex << ret << dec << emit();
	if (infiles) {
		info_summary_files_reached() <<
		 dataset::done_files() << " out of " << infiles <<
		 " input files were reached; "
		 << (infiles - dataset::done_files()) << " files were not reached"
		 << emit();
		info_summary_files_dropped() <<
			 (dataset::done_files() - dataset::error_files())
			 <<  " out of " << dataset::done_files() <<
			 " files reached were valid; " << dataset::error_files() <<
			 " were abandoned due to parse errors" << emit();
		if (infiles == dataset::done_files() && dataset::error_files() == 0) {
			if (line_despatch::lines_suppressed()) {
				info_summary_summary_dropped_lines() <<
					"Input lines were dropped" << emit();
			}
			if (line_despatch::lines_changed()) {
				info_summary_summary_changed_lines() <<
					 "Input lines were changed" << emit();
			}
			if (_error_directives_generated_) {
				warning_summary_summary_errored_lines() <<
					"Input lines were changed to #error directives"
					<< emit();
			}
			if (_error_directives_operative_) {
				warning_summary_summary_error_output() <<
				   "Unconditional #error directives are operative"
				   << emit();
			}
		}
	}
}

int diagnostic_base::exitcode()
{
	unsigned severities = 0;
	unsigned summaries = 0;
	if (line_despatch::lines_suppressed()) {
		summaries |= 16;
		++_infos_;
	}
	if (line_despatch::lines_changed()) {
		summaries |= 32;
		++_infos_;
	}
	if (_error_directives_generated_) {
		summaries |= 64;
		++_warnings_;
	}
	if (_error_directives_operative_) {
		summaries |= 128;
		++_warnings_;
	}
	if (_infos_) {
		severities |= 1;
	}
	if (_warnings_) {
		severities |= 2;
	}
	if (_errors_) {
		severities |= 4;
	}
	if (_abends_) {
		severities |= 8;
	}
	return severities < 4 ? severities | summaries : severities;
}

void diagnostic_base::help()
{
	cout << usage() << '\n';
	exit(EXIT_SUCCESS);
}


#ifndef PACKAGE_VERSION
#ifdef NIX
#error PACKAGE_VERSION not defined.
#endif
/*! Version of coan. Only used in Windows builds; otherwise defined via
    \c autoconf.
*/
#define PACKAGE_VERSION "5.1"
#endif
void diagnostic_base::version()
{
	cout << options::prog_name() << ", version " << PACKAGE_VERSION <<
	     " for " << OS_TYPE << "(built " <<  __DATE__ << ',' <<
	     __TIME__ << ')' << '\n';
	exit(EXIT_SUCCESS);
}

string diagnostic_base::usage()
{
	return  string("usage:\n")
	        + options::prog_name() + " COMMAND [OPTION... [FILE...]]\n"
	        "1. " + options::prog_name() + " { help | --help | -h }\n"
	        "\tDisplay this message and exit.\n"
	        "2. " + options::prog_name() + " { version | --version | -v }\n"
	        "\tDisplay version information and exit.\n"
	        "3. " + options::prog_name() + " source OPTION... [FILE...]\n"
	        "\tSimplify source FILEs in accordance with OPTIONs, "
	        "parsing in the manner of the C preprocessor.\n"
	        "\tInput lines may be dropped from output. Truth functional "
	        "directives may be simplified on output.\n"
	        "4. " + options::prog_name() + " spin OPTION... [FILE...]\n"
	        "\tLike the source command, with all simplified source files "
	        "output beneath a specified directory, mirroring their original "
	        "structure.\n"
	        "5. " + options::prog_name() + " symbols OPTION... [FILE...]\n"
	        "\tList preprocessor symbols from FILEs in accordance with "
	        "OPTIONs, parsing in the manner of the C preprocessor.\n"
	        "6. " + options::prog_name() + " includes OPTION... [FILE...]\n"
	        "\tList #includes fom FILEs in accordance with OPTIONs, parsing in "
	        "the manner of the C preprocessor\n"
	        "7. " + options::prog_name() + " defs OPTION... [FILE...]\n"
	        "\tList #defines and #undefs fom FILEs in accordance with OPTIONs, "
	        "parsing in the manner of the C preprocessor.\n"
	        "8. " + options::prog_name() + " pragmas OPTION... [FILE...]\n"
	        "\tList #pragmas fom FILEs in accordance with OPTIONs, parsing in "
	        "the manner of the C preprocessor\n"
	        "9. " + options::prog_name() + " directives OPTION... [FILE...]\n"
	        "\tList #-directives fom FILEs in accordance with OPTIONs, "
	        "parsing in the manner of the C preprocessor. "
	        "(Directives are #includes, #defines, #undefs, "
	        "#pragmas, #errors).\n"

	        "General OPTIONS:\n"
	        "\t-fARGFILE, --file ARGFILE\n"
	        "\t\tRead other arguments from ARGFILE.\n"
	        "\t-R, --recurse\n"
	        "\t\tRecurse into directories to find input files.\n"
	        "\t-FEXT1[,EXT2...]\n"
	        "\t--filter EXT1[,EXT2...]\n"
	        "\t\tProcess only input files that have one of the file extensions "
	        "EXT1,EXT2...\n"
	        "\t-DSYM[=VAL], --define SYM[=VAL]\n"
	        "\t\tAssume symbol SYM to be defined [=VAL]\n"
	        "\t-USYM, --undef SYM\n"
	        "\t\tAssume symbol SYM to be undefined.\n"
	        "\t-m, --implicit\n"
	        "\t\tAssume that any symbol that is not --define-ed is implicitly "
	        "--undef-ed.\n"
	        "\t-g{p|i|w|e|a},\n"
	        "\t--gag { progress| info | warning | error | abend }\n"
	        "\t\tSuppress diagnostics no worse than "
	        "{progress | info | warning | error | abend}.\n"
	        "\t-gs, --gag summary\n"
	        "\t\tSuppress summary diagnostics at exit.\n"
	        "\t\t(Default: -gp -gi -gs)\n"
	        "\t-V, --verbose\n"
	        "\t\tOutput all diagnostics.\n"
	        "\t-K, --keepgoing\n"
	        "\t\tIf a parse error is encountered in an input file, continue "
	        "processing subsequent input files.\n"
			"\t--no-transients\n"
			"\t\tBy default an in-source #define SYM or #undef SYM directive "
			"is transiently treated as a -DSYM or -USYM option within the "
			"source file where it is found. This option suppresses the "
			"default behaviour at your own risk.\n"
	        "\t-P, --pod\n"
	        "\t\tApart from #-directives, input is Plain Old Data.\n"
	        "\t\tComments and quotations will not be parsed.\n"
	        "\t-E, --evalconsts\n"
	        "\t\tEvaluate constants occurring as truth-functional "
	        "operands #ifs. "
	        "By default constants as truth-functional operands are not "
	        "evaluated.\n"

	        "source OPTIONs:-\n"
	        "\t-r, --replace\n"
	        "\t\tOverwrite input file with output file. Implied by --recurse\n"
	        "\t\tWith -r, stdin supplies the input *filenames*.\n"
	        "\t\tOtherwise stdin supplies an input *file*; "
	        "the output file is cout.\n"
	        "\t-bSUFFIX, --backup SUFFIX\n"
	        "\t\tBackup each input file by appending SUFFIX to the name.\n"
	        "\t\tApplies only with -r.\n"
	        "\t-kd, --discard drop\n"
	        "\t\tDrop discarded lines from output.\n"
	        "\t-kb, --discard blank\n"
	        "\t\tBlank discarded lines on output.\n"
	        "\t-kc, --discard comment\n"
	        "\t\tComment out discarded lines on output.\n"
	        "\t-xd, --conflict delete\n"
	        "\t\tDelete #defines and #undefs that contradict -D or "
	        "-U options.\n"
	        "\t-xc, --conflict comment\n"
	        "\t\tInsert diagnostic comments on contradictions as per -xd. "
	        "(Default.)\n"
	        "\t-xe, --conflict error\n"
	        "\t\tInsert diagnostic #errors on contradictions as per -xd.\n"
	        "\t--line\n"
	        "\t\tGenerate #line directives to make CPP line-numbering of output"
	        " agree with input even of lines are dropped.\n"
	        "\t-c, --complement\n"
	        "\t\tComplement. Retain the lines that ought to be dropped and "
	        "vice versa.\n"
	        "\t\t(Retained #-directives will still be simplified where "
	        "possible.)\n"

	        "spin OPTIONs:-\n"
	        "\tThe --recurse and --replace options are implied for this "
	        "command.\n"
	        "\t-kd, --discard drop\n"
	        "\t\tDrop discarded lines from output.\n"
	        "\t-kb, --discard blank\n"
	        "\t\tBlank discarded lines on output.\n"
	        "\t-kc, --discard comment\n"
	        "\t\tComment out discarded lines on output.\n"
	        "\t-xd, --conflict delete\n"
	        "\t\tDelete #defines and #undefs that contradict -D or "
	        "-U options.\n"
	        "\t-xc, --conflict comment\n"
	        "\t\tInsert diagnostic comments on contradictions as per -xd. "
	        "(Default.)\n"
	        "\t-xe, --conflict error\n"
	        "\t\tInsert diagnostic #errors on contradictions as per -xd.\n"
	        "\t--line\n"
	        "\t\tGenerate #line directives to make CPP line-numbering of output"
	        " agree with input even of lines are dropped.\n"
	        "\t-c, --complement\n"
	        "\t\tComplement. Retain the lines that ought to be dropped and "
	        "vice versa.\n"
	        "\t\t(Retained #-directives will still be simplified where "
	        "possible.)\n"
	        "\t--dir DIRNAME\n"
	        "\t\tGenerate a spin under directory DIRNAME, which will be "
	        " created if required. Every simplified source file will be output "
	        "to a file with the same name, relative to DIRNAME.\n"
	        "\t-pPATHNAME, --prefix PATHNAME\n"
	        "\t\tIn organizing output files beneath the directory DIRNAME, "
	        "DIRNAME will be equated to the path prefix PATHNAME of any "
	        "input file.\n"
	        "symbols OPTIONs:-\n"
	        "\t-i, --ifs\n"
	        "\t\tList symbols that occur #if[[n]def]/#elif directives.\n"
	        "\t-d, --defs\n"
	        "\t\tList symbols that occur in #define directives.\n"
	        "\t-u, --undefs\n"
	        "\t\tList symbols that occur in #undef directives.\n"
	        "\t--includes\n"
	        "\t\tList symbols that occur in #include directives.\n"
	        "\t(Default: -i -d -u --includes)\n"
	        "\t-o, --once\n"
	        "\t\tList only the first occurrence of each symbol. "
	        "(Default: List all occurrences.)\n"
	        "\t-A, --active\n"
	        "\t\tList only symbols from operative directives\n"
	        "\t-I, --inactive\n"
	        "\t\tList only symbols from inoperative directives\n"
	        "\t-L, --locate\n"
	        "\t\tReport the source file and line number of each listed "
	        "occurrence.\n"
	        "\t-e, --evalsyms\n"
	        "\t\tReport the resolved value of each symbol.\n"
	        "includes OPTIONs:-\n"
	        "\t-s, --system\n"
	        "\t\tList system headers, i.e. <header.h>\n"
	        "\t-L, --local\n"
	        "\t\tList local headers, i.e. \"header.h\"\n"
	        "\t(Default: -s -L)\n"
	        "\t-o, --once\n"
	        "\t\tList only the first occurrence of each header. "
	        "(Default: List all occurrences.)\n"
	        "\t-A --active\n"
	        "\t\tList only headers from operative directives.\n"
	        "\t-I --inactive\n"
	        "\t\tList only headers from inoperative directives.\n"
	        "\t-L, --locate\n"
	        "\t\tReport the source file and line number of each listed "
	        "occurrence.\n"

	        "defs OPTIONs:-\n"
	        "\t-o, --once\n"
	        "\t\tList only the first occurrence of each distinct #define or "
	        "#undef directive (Default: List all occurrences.)\n"
	        "\t-A --active\n"
	        "\t\tList only operative #define and #undef directives.\n"
	        "\t-I --inactive\n"
	        "\t\tList only inoperative #define and #undef directives.\n"
	        "\t-L, --locate\n"
	        "\t\tReport the source file and line number of each listed "
	        "occurrence.\n"

	        "pragmas OPTIONs:-\n"
	        "\t-o, --once\n"
	        "\t\tList only the first occurrence of each distinct pragma "
	        "(Default: List all occurrences.)\n"
	        "\t-A, --active\n"
	        "\t\tList only pragmas from lines that are retained.\n"
	        "\t-I, --inactive\n"
	        "\t\tList only pragmas from lines that are dropped.\n"
	        "\t-L, --locate\n"
	        "\t\tReport the source file and line number of each listed "
	        "occurrence.\n"

	        "errors OPTIONs:-\n"
	        "\t-o, --once\n"
	        "\t\tList only the first occurrence of each #error directive "
	        "(Default: List all occurrences.)\n"
	        "\t-A, --active\n"
	        "\t\tList only operative #error directives.\n"
	        "\t-I, --inactive\n"
	        "\t\tList only inoperative #error directives.\n"
	        "\t-L, --locate\n"
	        "\t\tReport the source file and line number of each listed "
	        "occurrence.\n"
	        "directives OPTIONs:-\n"
	        "\t-o, --once\n"
	        "\t\tList only the first occurrence of each distinct directive "
	        "(Default: List all occurrences.)\n"
	        "\t-A, --active\n"
	        "\t\tList only operative directives.\n"
	        "\t-I, --inactive\n"
	        "\t\tList only inoperative directives.\n"
	        "\t-L, --locate\n"
	        "\t\tReport the source file and line number of each listed "
	        "occurrence.";
}

string diagnostic_base::citable_string(std::string const & str)
{
	string citable;
	size_t len(str.length());
	bool escaped = false;
	size_t ndx = 0;
	char const space = ' ';
	char last_ch = 0;
	for (   ; ndx < len; ++ndx) {
		char ch(str[ndx]);
		switch(ch) {
		case '\\':
			if (!escaped) {
				if (ndx + 2 < len &&
				    str[ndx + 1] == '\r' &&
				    str[ndx + 2] == '\n') {
					if (last_ch != space) {
						citable.append(1,space);
						last_ch = space;
					}
					ndx += 2;
				} else if (ndx + 1 < len && str[ndx + 1] == '\n') {
					if (last_ch != space) {
						citable.append(1,' ');
						last_ch = space;
					}
					++ndx;
				} else {
					citable.append(1,ch);
					last_ch = ch;
					escaped = true;
				}
				continue;
			}
			break;
		case '\r':
			if (ndx + 1 < len &&
			    str[ndx + 1] == '\n') {
				++ndx;
			}
			if (last_ch != space) {
				citable.append(1,space);
				last_ch = space;
			}
			break;
		case '\n':
		case '\t':
			if (last_ch != space) {
				citable.append(1,space);
				last_ch = space;
			}
			break;
		default:
			if (isprint(ch)) {
				citable.append(1,ch);
				last_ch = ch;
			}
		}
		escaped = false;
	}
	ndx = citable.length();
	if (ndx) {
		// Trim trailing space
		while (ndx-- > 0 && isspace(citable[ndx])) {}
		citable.resize(ndx + 1);
		size_t len = citable.length();
		// Trim leading space
		for (ndx = 0; ndx < len && ndx == space; ++ndx);
		if (ndx > 0) {
			citable.erase(0,ndx);
		}
	}
	return citable;
}

string diagnostic_base::citable_string(integer const & integ)
{
	stringstream numeral;
	switch(integ.type()) {
	case INT_INT:
		numeral << (int)integ.raw();
		break;
	case INT_UINT:
		numeral << (unsigned)integ.raw();
		break;
	case INT_LONG:
		numeral << (long)integ.raw();
		break;
	case INT_ULONG:
		numeral << (unsigned long)integ.raw();
		break;
	case INT_LLONG:
		numeral << (long long)integ.raw();
		break;
	case INT_ULLONG:
		numeral << (unsigned long long)integ.raw();
		break;
	default:
		assert(false);
	}
	return numeral.str();
}

string diagnostic_base::citable_string(char const *start, char const *end)
{
	string result;
	for (   ; start < end; start = chewer<continuation>(++start)) {
		if (isprint(*start)) {
			result += *start;
		}
	}
	return result;
}

size_t
diagnostic_base::volley(unsigned code,
                        void (diagnostic_base::*method)() const)
{
	size_t shots = 0;
	list<ptr>::iterator qit(_queue_.begin());
	while ( qit != _queue_.end()) {
		if (qit->get()->encode() == code) {
			if (method) {
				(qit->get()->*method)();
			}
			qit = _queue_.erase(qit);
			++shots;
		} else {
			++qit;
		}
	}
	return shots;
}

void diagnostic_base::flush_all()
{
	for(    ; !_queue_.empty(); _queue_.pop_front()) {
		_queue_.begin()->get()->despatch();
	}
}

template<char Delim>
string system_error_message(unsigned errnum);

#if PATH_DELIM == '/'
#include <cstring>

template<>
string system_error_message<'/'>(unsigned errnum)
{
	return strerror(errnum);
}

#else
#include <sstream>
#include <windows.h>

template<>
string system_error_message<'\\'>(unsigned errnum)
{
	std::ostringstream gripe;
	LPTSTR mess_buf = 0;
	DWORD chars = FormatMessage(
	                  FORMAT_MESSAGE_ALLOCATE_BUFFER |
						FORMAT_MESSAGE_FROM_SYSTEM,
	                  NULL,
	                  errnum,
	                  LANG_USER_DEFAULT,
	                  (LPTSTR)&mess_buf,
	                  0,
	                  NULL);
	if (!chars) {
		gripe << "system error " << errnum;
	} else {
		gripe << mess_buf;
	}
	if (mess_buf) {
		LocalFree(mess_buf);
	}
	return gripe.str();
}

#endif

std::string diagnostic_base::system_error_message(unsigned errnum)
{
	return ::system_error_message<PATH_DELIM>(errnum);
}

// EOF
