#ifndef DIAGNOSTIC_H
#define DIAGNOSTIC_H

/***************************************************************************
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "platform.h"
#include "integer.h"
#include <sstream>
#include <list>
#include <vector>
#include <map>
#include <memory>

/*! \file diagnostic.h
    This file defines class \e diagnostic
*/

//! Class \c severity is a namespace for its enumerated member constants
struct severity {
	//! Enumerated constants representing the severities of diagnostics
	enum level {
	    //! No severity
	    none = 0,
	    //! A progress diagnostic
	    progress = 1,
	    //! An informational diagnostic
	    info = 2,
	    //! A warning diagnostic
	    warning = 4,
	    //! An error diagnostic
	    error = 8,
	    //! A fatal error disgnostic
	    abend = 16,
	    //! A summary diagnostic
	    summary = 32,
	    //! A progress summary
	    summary_progress = summary | progress,
	    //! An informational summary
	    summary_info = summary | info,
	    //! A warning summary
	    summary_warning = summary | warning
	};
};

/*! \brief The token class is inserted in a \c diagnostic_base to tell it to
    emit itself
*/
struct emit {};

/*! \brief The token class is inserted in a \c diagnostic_base to tell it to
    defer itself
*/
struct defer {};

//! A base class for diagnostic classes
struct diagnostic_base {
#if CPP11_HAVE_MATURE_UNIQUE_PTR
	//! Type of pointer to a \c diagnostic_base
	typedef std::unique_ptr<diagnostic_base> ptr;
#else
	typedef std::shared_ptr<diagnostic_base> ptr;
#endif

	//! Destructor
	virtual ~diagnostic_base() {}

	//! Globally enqueue a copy of this \c diagnostic_base for deferred action.
	void queue() const {
		_queue_.push_back(ptr(clone()));
	}

	//! Get the severity level of the runtime type.
	virtual severity::level level() const {
		return severity::none;
	}

	/*! \brief Get the unique identifier of the runtime type.

	    Each distinct cause that may be reported by a diagnostic has a
	    distinct identifier.

	    \return The unique identifier of the runtime type.
	*/
	virtual unsigned id() const {
		return 0;
	}

	/*! \brief Get the reason-code of the diagnostic.

	    The reason-code encodes the id of the diagnostic and its
	    severity level.

	    \return The reason-code of \c diagnostic.
	*/
	unsigned encode() const {
		return code(level(),id());
	}

	//! Cast the object to its reason-code.
	operator unsigned () const {
		return encode();
	}

	//! Say whether this diagnostic is suppressed.
	bool gagged() const {
		return _gagged;
	}

	//! Get the text of the diagnostic
	std::string text() const {
		return _text.str();
	}

	//! Cast the object to its textual representation.
	operator std::string const () {
		return text();
	}

	//! Count the diagnostics enqueued for deffered action.
	static size_t deferred() {
		return _queue_.size();
	}

	/*! \brief Discard selected queued diagnostics.

	    Discard any queued diagnostics that match a
	    reason-code, returning the number discarded.
	    \param  encode The reason-code for selecting the queued diagnostics to
				discard.
	    \return The number of discarded diagnostics.
	*/
	static size_t discard(unsigned encode) {
		return volley(encode);
	}
	
	//! Discard all queued diagnostics.
	static void discard_all() {
		_queue_.resize(0);
	}

	//! Emit all queued diagnostics that match a reason-code
	static size_t flush(unsigned encode) {
		return volley(encode,&diagnostic_base::despatch);
	}

	/*! Emit any queued diagnostics */
	static void flush_all();

	/*! \brief Write summary diagnostics on \c cerr at exit.

	    The summary diagnostics will be output unless \c --gag \c summary
	    is in force, which it is by default.
	*/
	static void epilogue();

	/*! \brief Convert the internal exit status to an exit code for \c exit().

	    \return RC such that:

	    - RC = 0:			MSG_CLASS_none: No diagnostics accrued.

	    - RC & 0x2:		MSG_CLASS_info: 	At worst info diagnostics accrued.

	    - RC & 0x4:		MSG_CLASS_warning: At worst warnings accrued.

	    - RC & 0x8:		MSG_CLASS_error:	At worst errors accrued.

	    - RC & 0x16:		MSG_CLASS_abend:	Execution was aborted.

	    - RC & 0x12:		EVENT_summary_DROPPED_LINES (info): Input lines were
	                        dropped on output.

	    - RC & 0x22:	EVENT_summary_CHANGED_LINES (info): Input lines were
	                    changed on  output.

	    - RC & 0x44:	EVENT_summary_errorED_LINES (warning): Lines were
	                    changed to \c hash-errors on output.

	    - RC & 0x84:	EVENT_summary_error_OUPUT (warning): Unconditional
	                    hash-errors were  output.
	*/
	static int exitcode();

	/*! \brief Enter or leave code-parsing mode, returning the out-going mode.
	    \param  enable  If true, enable code-parsing mode, else diable.
	    \return The previous status of code-parsing mode.
	*/
	static bool code_mode(bool enable = true) {
		bool old_mode = _in_code_;
		_in_code_ = enable;
		return old_mode;
	}
	//! Say whether we are in code-parsing mode.
	static bool in_code_mode() {
		return _in_code_;
	}

	/*! \brief Make a citable version of a string.

	    Make a citable version of a string - i.e. suitable for
	    quoting in a diagnostic - by stripping out linefeeds and
	    replacing tabs with spaces.
	    \param  str The string to be converted.
	    \return A citable version of \c str.
	*/
	static std::string citable_string(std::string const & str);

	/*! \brief Make a citable version of an integral object.
	    Make a citable version of an integral object - i.e. suitable for
	    quoting in a diagnostic.
	    \param  n   The \c integer to be converted to a citable string.
	    \return A citable string representation of \c n.
	*/
	static std::string citable_string(integer const & n);

	/*! \brief Make a citable version of delimited text.
	    Make a citable version of text between start and end points,
	    in which non-printable characters may occur.
	    \param  start   The start of the text to scan.
	    \param  end     The end of the text to scan.
	    \return A string extracted from the delimited span by deleting
	    any line-continuations non-printable characters.
	*/
	static std::string citable_string(char const *start, char const *end);



	/*!	\brief Write brief command usage information to \c cout.

	    The usage info is written the \c --help option,
	    then the program is terminated without error.
	*/
	static void help();

	//!	Write version information about the program on \c cout.
	static void version();

	//! Get program usage summary in a string.
	static std::string usage();

	/*! \brief Get the system error message for a system error number.
	    \param  errnum The system error number for which the message is required.
	    \return The system error message associated with \c errnum.
	*/
	static std::string system_error_message(unsigned errnum);

protected:

	/*! \brief Constructor.

	    \param severity_name Name of the severity level of the diagnostic
	    \param  level   The severity level of the diagnostic
	    \param  id      The causal identifier of the diagnostic
	    \param  in_code_only True if the diagnostic is applicable only when
	            parsing code.

	*/
	diagnostic_base(std::string severity_name,
	                severity::level level, unsigned id,
	                bool in_code_only = false);

	//! Copy constructor.
	diagnostic_base(diagnostic_base const & src)
		: _gagged(src._gagged),_text(src._text.str()) {}

#if CPP11_HAVE_DECL_DELETE
	//! Assign a \c diagnostic_base from another.
	diagnostic_base & operator=(diagnostic_base other) = delete;
#else
	diagnostic_base & operator=(diagnostic_base other);
#endif

	//! Increment global counts appropriately for the runtime type.
	virtual void count() const {}

	//! Clone the runtime object on the heap.
	virtual diagnostic_base * clone() const {
		return new diagnostic_base(*this);
	}

	//! Emit the diagnostic
	void despatch() const throw (unsigned);

	//! Is the diagnostic suppressed?
	bool _gagged;
	//! The text of the diagnostic.
	std::stringstream _text;

	/*! \brief Get the reason-code that encodes a causal identifier and severity
	    level.

	    \param  level   The severity level to be encoded.
	    \param  id  The causal identifier to be encoded.
	    \return The reason-code that encodes a causal identifier and severity
	    level.
	*/
	static unsigned code(severity::level level, unsigned id) {
		return (level << 8) | id;
	}
	/*! \brief Queue of deferred diagnostics.

	    A diagnostic may have to be constructed before we know whether it
	    should be issued. In the meantime it can be queued.
	*/
	static std::list<ptr> _queue_;

	//! Are we parsing C/C++ code?
	static bool _in_code_;
	//! Global count of informational diagnostics.
	static unsigned _infos_;
	//! Global count of warning diagnostics.
	static unsigned _warnings_;
	//! Global count of error diagnostics.
	static unsigned _errors_;
	//! Global count of fatal error diagnostics.
	static unsigned _abends_;
	//! Global count of \#error directives output.
	static unsigned _error_directives_generated_;
	//! Global count of operative \#error directives output.
	static unsigned _error_directives_operative_;

private:

	/*! \brief Call a member function on all queued diagnostics for a given
	    reason-code.
	    \param  encode  The reason-code for which to select queued diagnostics.
	    \param  method The member method to be called on selected diagnostics.
	    \return The number of diagnostics selected.
	*/
	static size_t
	volley(unsigned encode, void (diagnostic_base::*method)() const = NULL);
};

/*! \brief Class template \c diagnostic<Level,Id,InCodeOnly>
    encapsulates a type of diagnostic.

    \tparam  Level   The severity level of the diagnostic type.
    \tparam Id The causal identifier of diagnostic type.
    \tparam  InCodeOnly  True denotes the diagnostic type is applicable
        only when parsing code.
*/
template<severity::level Level, unsigned Id, bool InCodeOnly = false>
struct diagnostic : diagnostic_base {
	//! The severity level of this diagnostic type.
	static const severity::level _severity_ = Level;
	//! The causal identifier of this diagnostic type.
	static const unsigned _id_ = Id;

	/*! \brief Constructor.

	    Explicitly construct a \c diagnostic given a severity name.
	    \param  severity_name Severity name of the diagnostic to be
	        constructed.
	*/
	explicit diagnostic(std::string const & severity_name)
		: diagnostic_base(severity_name,Level,Id,InCodeOnly) {}

	//! Destructor
	virtual ~diagnostic() {};

	//! Get the severity level of this \c diagnostic.
	severity::level level() const {
		return Level;
	}

	//! Get the causal identifier of this \c diagnostic.
	unsigned id() const {
		return Id;
	}

	//! Clone this \c diagnostic on the heap.
	diagnostic_base * clone() const {
		return new diagnostic(*this);
	}

	/*! \brief Insert an arbitrary object into this \c diagnostic.
	    \tparam T Type of the object to be inserted.
	    \param  obj The object to be inserted.
	    \return A reference to \c this.

	    The member function inserts \c obj into the text
	    of the diagnostic.
	*/
	template<typename T>
	diagnostic & operator<<(T const & obj) {
		if (!_gagged) {
			_text << obj;
		}
		return *this;
	}

	/*! \brief Queue the diagnostic for deferred action.

	    Inserting a token object of type \c defer to the \c diagnostic
	    adds the diagnostic to the deferred queue.

	    \return A reference to \c this.
	*/
	diagnostic & operator<<(defer) {
		if (!_gagged) {
			queue();
		}
		return *this;
	}

	/*! \brief Emit the diagnostic.

	    Inserting a token object of type \c emit to the \c diagnostic
	    causes it to be emitted.

	    \return A reference to \c this.
	*/
	diagnostic & operator<<(emit) {
		if (!_gagged) {
			despatch();
		}
		return *this;
	}

	//! Get the reason-code of this \c diagnostic's type
	static unsigned code() {
		return diagnostic_base::code(Level,Id);
	}

	//! Forget this \c diagnostic.
	static size_t discard() {
		return diagnostic_base::discard(code());
	}

	//! Emit all deferred diagnostics of this type.
	static size_t flush() {
		return diagnostic_base::flush(code());
	}
};

/*! \brief Class \c progress_msg encapsulates a progress diagnostic.
    \tparam Id The causal identifier of \c progress_msg type.
    \tparam  InCodeOnly  True denotes the \c progress_msg type is applicable
        only when parsing code.
*/
template<unsigned Id, bool InCodeOnly = false>
struct progress_msg : diagnostic<severity::progress,Id,InCodeOnly> {
	//! Constructor
	progress_msg()
		: diagnostic<severity::progress,Id,InCodeOnly>("progress") {}

	//! \brief Clone this \c progress_msg on the heap.
	diagnostic_base * clone() const {
		return new progress_msg(*this);
	}
};

/*! \brief Class \c info_msg encapsulates an informational diagnostic.
    \tparam Id The causal identifier of \c info_msg type.
    \tparam  InCodeOnly  True denotes the \c info_msg type is applicable
        only when parsing code.
*/
template<unsigned Id, bool InCodeOnly = false>
struct info_msg : diagnostic<severity::info,Id,InCodeOnly> {
	using diagnostic<severity::info,Id,InCodeOnly>::_infos_;

	//! Default constructor
	info_msg()
		: diagnostic<severity::info,Id,InCodeOnly>("info") {}

	//! Clone this \c info_msg on the heap.
	diagnostic_base * clone() const {
		return new info_msg(*this);
	}

private:

	//! Increment the global count of informational diagnostics
	void count() const {
		++_infos_;
	};

};

/*! \brief Class \c warning_msg encapsulates a warning diagnostic.
    \tparam Id The causal identifier of \c warning_msg type.
    \tparam  InCodeOnly  True denotes the \c warning_msg type is applicable
        only when parsing code.
*/
template<unsigned Id, bool InCodeOnly = false>
struct warning_msg : diagnostic<severity::warning,Id,InCodeOnly> {
	using diagnostic<severity::warning,Id,InCodeOnly>::_warnings_;

	//! Default constructor
	warning_msg()
		: diagnostic<severity::warning,Id,InCodeOnly>("warning") {}
	~warning_msg() {}

	//! Clone this \c warning_msg on the heap.
	diagnostic_base * clone() const {
		return new warning_msg(*this);
	}
private:
	//! Increment the global count of warning diagnostics
	virtual void count() const {
		++_warnings_;
	};

};

/*! \brief Class \c error_msg encapsulates a error diagnostic.
    \tparam Id The causal identifier of \c error_msg type.
    \tparam  InCodeOnly  True denotes the \c error_msg type is applicable
        only when parsing code.
*/
template<unsigned Id, bool InCodeOnly = false>
struct error_msg : diagnostic<severity::error,Id,InCodeOnly> {
	using diagnostic<severity::error,Id,InCodeOnly>::_errors_;

	//! Default constructor
	error_msg()
		: diagnostic<severity::error,Id,InCodeOnly>("error") {}

	//! Clone this \c error_msg on the heap.
	diagnostic_base * clone() const {
		return new error_msg(*this);
	}
private:
	//! Increment the global count of error diagnostics
	void count() const {
		++_errors_;
	};

};

/*! \brief Class \c abend_msg encapsulates a fatal error diagnostic.

    \tparam Id The causal identifier of \c abend_msg type.
    \tparam  InCodeOnly  True denotes the \c abend_msg type is applicable
        only when parsing code.
*/
template<unsigned Id, bool InCodeOnly = false>
struct abend_msg : diagnostic<severity::abend,Id,InCodeOnly> {
	using diagnostic<severity::abend,Id,InCodeOnly>::_abends_;

	//! Default constructor
	abend_msg()
		: diagnostic<severity::abend,Id,InCodeOnly>("abend") {}

	//! Clone this \c abend_msg on the heap.
	diagnostic_base * clone() const {
		return new abend_msg(*this);
	}
private:
	//! Increment the global count of abend diagnostics
	void count() const {
		++_abends_;
	};

};

/*! \brief Class \c progress_summary_msg encapsulates a progress summary
    diagnostic.

    \tparam Id The causal identifier of \c progress_summary_msg type.
    \tparam  InCodeOnly  True denotes the \c progress_summary_msg type is
        applicable only when parsing code.
*/
template<unsigned Id, bool InCodeOnly = false>
struct progress_summary_msg :
		diagnostic<severity::summary_progress,Id,InCodeOnly> {
	//! Default constructor
	progress_summary_msg()
		: diagnostic<severity::summary_progress,Id,InCodeOnly>("progress") {}
};

/*! \brief Class \c info_summary_msg encapsulates an informational summary
    diagnostic.

    \tparam Id The causal identifier of \c info_summary_msg type.
    \tparam  InCodeOnly  True denotes the \c info_summary_msg type is
        applicable only when parsing code.
*/
template<unsigned Id, bool InCodeOnly = false>
struct info_summary_msg : diagnostic<severity::summary_info,Id,InCodeOnly> {
	//! Default constructor
	info_summary_msg()
		: diagnostic<severity::summary_info,Id,InCodeOnly>("info") {}
};

/*! \brief Class \c warning_summary_msg encapsulates a warning summary
    diagnostic.

    \tparam Id The causal identifier of \c warning_summary_msg type.
    \tparam  InCodeOnly  True denotes the \c warning_summary_msg type is
        applicable only when parsing code.
*/
template<unsigned Id, bool InCodeOnly = false>
struct warning_summary_msg :
		diagnostic<severity::summary_warning,Id,InCodeOnly> {
	//! Default constructor
	warning_summary_msg()
		: diagnostic<severity::summary_warning,Id,InCodeOnly>("warning") {}
};

//! Report file being processed.
typedef progress_msg<1> progress_processing_file;
//! Peport building the input tree.
typedef progress_msg<2> progress_building_tree;
//! Report added a file to input.
typedef progress_msg<3> progress_added_file;
//! Report total input files found.
typedef progress_msg<4> progress_file_tracker;
//! Report a commandline argument parsed.
typedef progress_msg<5> progress_got_options;

//! Report a duplicate diagnostic selection option
typedef info_msg<1> info_duplicate_mask;
//! Report that input file or directory is symbolic link.
typedef info_msg<2> info_symlink;

/*! \brief Report that same argument occurs for multiple \c --define or
	\c --undefine options
*/
typedef warning_msg<1> warning_duplicate_args;

/*! \brief Report that an input \c \#define or \c \#undef directive was deleted.

    Report that an input \c \#define or \c \#undef directive was deleted on
    output because it contradicts a \c --define or \c --undefine
	option together with \c --contradict \c delete
*/
typedef warning_msg<2> warning_deleted_contradiction;
/*!
    \brief Report that an input \c \#define or \c \#undef directive was
    commented out.

    Report that an input \c \#define or \c \#undef directive was commented out
    on output because it contradicts a \c --define or \c --undefine
	option togther with \c --contradict \c comment
*/
typedef warning_msg<3> warning_commented_contradiction;
/*!

    \brief Report that an input \c \#define or \c \#undef directive was
    converted to an \c \#error.

    Report that an input \c \#define or \c \#undef directive was converted to a
	\c \#error on output because it contradicts a \c --define or \c --undefine
	option \c together with \c --contradict \c error
*/
typedef warning_msg<4> warning_errored_contradiction;
/*! \brief Report that an in-source \c \#define defines a symbol differently
	from a prior one.
*/
typedef warning_msg<5> warning_differing_redef;

/*! \brief Report that an in-source \c \#undef undefines a symbol previously
	defined by an in-source \c \#define.
*/
typedef warning_msg<6> warning_undefing_defined;

/*! \brief Encapsulates a diagnostic for \c \#error directive input
    unconditionally.

    Class \c warning_unconditional_error_input encapsulates a
    diagnostic to report that an \c \#error directive was input unconditionally,
    i.e. not in the scope of \c \#if
*/
struct warning_unconditional_error_input : warning_msg<7> {
private:
	//! Increment global counts appropriately.
	void count() const {
		++_warnings_;
		++_error_directives_operative_;
	}
};
/*! \brief Encapsulates a diagnostic for \c \#error directive output
        conditionally.

    Class \c warning_error_generated_input encapsulates a
    diagnostic to report that an \c \#error directive was output conditionally,
    i.e. within the scope of an \c \#if
*/
struct warning_error_generated : warning_msg<8> {
private:
	//! Increment global counts appropriately.
	void count() const {
		++_warnings_;
		++_error_directives_generated_;
	}
};
/*! \brief Encapsulates a diagnostic for \c \#error directive output
        unconditionally.

    Class \c warning_unconditional_error_output encapsulates a
    diagnostic to report that an \c \#error directive was output
    unconditionally, i.e. not in the scope of any \c \#if
*/
struct warning_unconditional_error_output : warning_msg<9> {
private:
	//! Increment global counts appropriately.
	void count() const {
		++_warnings_;
		++_error_directives_operative_;
	}
};
//! Report garbage text was input following a directive
typedef warning_msg<10> warning_garbage_after_directive;
//! Report a newline is missing at end of input*/
typedef warning_msg<11> warning_missing_eof_newline;
//! Report a symbol has a circular definition.
typedef warning_msg<12> warning_circular_sym_def;
//! Report the \c --verbose option is mixed with the \c --gag option
typedef warning_msg<13> warning_verbose_only;
//! Report a divide by zero was found in an expression
typedef warning_msg<14> warning_zero_divide;
/*! \brief Report directory name ignored on input when \c --recurse not
	not specified
*/
typedef warning_msg<15> warning_dir_ignored;

/*! \brief Report a symbol that evaluates to an empty string is an
	operand in an expression
*/
typedef warning_msg<16> warning_empty_symbol;
//! Report an unknown preprocessor directive
typedef warning_msg<17> warning_unknown_directive;
//! Report a problematic argument to an \c \#include directive
typedef warning_msg<18> warning_invalid_include;
//! Report a misplaced escape character '\'
typedef warning_msg<19,true> warning_stray_escape;
//! Report that an integer constant evaluates > \c INT_MAX
typedef warning_msg<20> warning_int_overflow;
//! Report unclosed quotation
typedef warning_msg<21,true> warning_unclosed_quote;
//! Report negative shift count
typedef warning_msg<22> warning_negative_shift;
//! Report a huge integer constant forced to be unsigned
typedef warning_msg<23> warning_forced_unsigned;
//! Report integer sign changed by promotion
typedef warning_msg<24> warning_sign_changed;
//! Report a that shift count is >= the width of the shifted quantity
typedef warning_msg<25> warning_shift_overflow;
//! Report a character constant too long for current locale
typedef warning_msg<26> warning_char_constant_too_long;
//! Report that character constant goes multi-byte
typedef warning_msg<27> warning_mulitbyte_char_constant;
//! Apologize that variadic macro expansion is unsupported.
typedef warning_msg<28> warning_variadic_macros_unsupported;
/*! Report that a symbol has been transiently defined or undefined for the
    duration of the current source file
*/
typedef warning_msg<28> warning_transient_symbol_added;
//! Report that the commandline does not specify any \c --define or \c --undef
typedef warning_msg<29> warning_no_syms;
//! Report that input file is a broken symbolic link
typedef warning_msg<30> warning_broken_symlink;
//! Report that the unsafe \c --no-transients option is used.
typedef warning_msg<31> warning_no_transients_used;

//! Report an orphan \c \#elif was found in input
typedef error_msg<1> error_orphan_elif;
//! Report an orphan \c \#else was found in input
typedef error_msg<2> error_orphan_else;
//! Report an orphan \c \#endif was found in input
typedef error_msg<3> error_orphan_endif;
//! Report unexpected end of input
typedef error_msg<4> error_eof_too_soon;
//! Report input ends within quotation
typedef error_msg<5> error_eof_in_quote;
//! Report that the commandline does not specify anything to do
typedef error_msg<6> error_nothing_to_do;
/*! Report that the commandline invalidly specifies multiple input files without
	the \c --replace option
*/
typedef error_msg<7> error_one_file_only;
//! Report that a newline occurs within quotation in input
typedef error_msg<8> error_newline_in_quote;
//! Report that the commandline options are not a valid combination
typedef error_msg<9> error_invalid_args;
//! Report that the commandline is syntactically invalid
typedef error_msg<10> error_usage;
//! Report that \c \#if nesting exceeds \c maxdepth in input
typedef error_msg<11> error_too_deep;
//! Report that the \c --file option occurs more than once
typedef error_msg<12> error_multiple_argfiles;
/*! Report that the argument to a \c --define or \c --undefine option
	contains mysterious characters.
*/
typedef error_msg<13> error_garbage_arg;
//! Report that a valid preprocessor identifier was not found where required
typedef error_msg<14> error_not_identifier;
//! Report unbalanced parenthesis
typedef error_msg<15> error_unbalanced_paren;
//! Report \c \#if or \c \#elif has no argument
typedef error_msg<16> error_if_without_cond;
//! Report failure from \c realpath()
typedef error_msg<17> error_cant_get_realpath;
//! Report a malformed macro parameter list
typedef error_msg<18> error_malformed_macro;
//! Report garbage in a directive
typedef error_msg<19> error_garbage_in_directive;
//! Report wrong number of macro arguments for symbol
typedef error_msg<20> error_mismatched_macro_args;
//! Report attempting macro-call on undefined symbol
typedef error_msg<21> error_macro_call_on_undefined;
//! Report an incomplete ternary conditional operator
typedef error_msg<22> error_ternary_cond_incomplete;
//! Report input ends within a C-comment
typedef error_msg<31> error_eof_in_comment;

//! Report cannot open an output file
typedef abend_msg<1> abend_cant_open_output;
//! Report cannot open an input file
typedef abend_msg<2> abend_cant_open_input;
/*! \brief Report end of input encountered while reading a quoted filename.
	a closing quotation was not found
*/
typedef abend_msg<3> abend_eof_in_filename;
//! Report an internal logic abend
typedef abend_msg<4> abend_confused;
//! Report read error on input
typedef abend_msg<5> abend_cant_read_input;
//! Report write error on output
typedef abend_msg<6> abend_cant_write_file;
//! Report can't identify file or directory
typedef abend_msg<7> abend_no_file;
//! Report invalid filename input
typedef abend_msg<8> abend_illegal_filename;
//! Report failure to delete a file
typedef abend_msg<9> abend_cant_delete_file;
//! Report failure to rename a file
typedef abend_msg<10> abend_cant_rename_file;
//! Report cannot create a temporary file
typedef abend_msg<11> abend_no_tempfile;
//! Report cannot open directory
typedef abend_msg<12> abend_cant_open_dir;
//! Report read error on directory
typedef abend_msg<13> abend_cant_read_dir;
/*! \brief Report that a spin directory includes, is included by or is same as
    an input directory
*/
typedef abend_msg<14> abend_invalid_spin_dir;
//! Report can't get the current working directory
typedef abend_msg<15> abend_cant_get_cwd;
//! Report can't create directory
typedef abend_msg<23> abend_cant_create_dir;


//! Report processing complete
typedef progress_summary_msg<1> progress_summary_all_done;
//! Report total files reached
typedef info_summary_msg<1> info_summary_files_reached;
//! Report total files abandoned due to errors
typedef info_summary_msg<2> info_summary_files_dropped;
//! Report that input lines were dropped
typedef info_summary_msg<3> info_summary_summary_dropped_lines;
//! Report that input lines were changed
typedef info_summary_msg<4> info_summary_summary_changed_lines;
//! Report that input lines were converted to \c \#errors
typedef warning_summary_msg<1> warning_summary_summary_errored_lines;
//! Report that unconditional \c \#error lines were output
typedef warning_summary_msg<2> warning_summary_summary_error_output;

#endif // EOF
