#ifndef DIRECTIVE_H
#define DIRECTIVE_H
/***************************************************************************
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "platform.h"
#include "line_despatch.h"
#include <map>
#include <vector>
#include <string>

/*!
    \file directive.h
    This defines class \e directive
*/


//! Symbolic constants denoting types of directives.
enum directive_type {
    HASH_UNKNOWN,	//!< An unknown directive
    HASH_IF,		//!< An \c \#if directive
    HASH_IFDEF,		//!< An \c \#ifdef directive
    HASH_IFNDEF,	//!< An \c \#ifndef directive
    HASH_ELSE,		//!< An \c \#else directive
    HASH_ELIF,		//!< An \c \#elif directive
    HASH_ENDIF,		//!< An \c \#endif directive
    HASH_DEFINE,	//!< An \c \#define directive
    HASH_UNDEF,		//!< An \c \#undef directive
    HASH_INCLUDE,	//!< An \c \#include directive
    HASH_PRAGMA,	//!< A \c \#pragma directive
    HASH_ERROR		//!< An \c \#error directive
};

/*! \brief Class \e directive_base is a base for instantiations of class
    template \c directive<Type>
*/
struct directive_base {
#if CPP11_HAVE_DECL_DELETE
	//! Copying a \c directive_base is not allowed
	directive_base(directive_base const &) = delete;
	//! Assigning a \c directive_base is not allowed
	directive_base & operator=(directive_base const &) = delete;
#else
	directive_base(directive_base const &);
	//! Assigning a \c directive_base is not allowed
	directive_base & operator=(directive_base const &);
#endif

	/*! \brief Evaluate a directive
	    \param  keyword The keyword indicating the directive type.
	    \param  ptext   On entry, a pointer to starting address of the
	        directive's body. On return, receives the first address not consumed
	        by scanning the directive.
	    \return A \c line_type determined by evaluation
	        denoting the action to be taken on this line.
	*/
	static line_type eval(std::string keyword, char const **ptext) {
		return eval(get_type(keyword),ptext);
	}

protected:

	//! Default constructor
	directive_base() {};
	/*! \brief Report a directive
	    \param  keyword The keyword indicating the directive type.
	    \param  arg The body of the directive.
	*/
	static void report(std::string const & keyword, std::string const & arg);

	/*! \brief Report a directive.
	    \param  seen Has this directive already been seen?
	    \param  keyword The keyword indicating the directive type.
	    \param  arg The body of the directive.

	    The member function reports the directive if it has not already been
	    seen or if option \c --once is not in effect.
	*/
	static void report(bool seen,
	                   std::string const & keyword, std::string const & arg);

	/*! \brief Evaluate a directive.
	    \param  type    The \c directive_type of the directive.
	    \param  ptext On entry, a pointer to starting address of the directive's
	        body. On return, receives the first address not consumed by
	        scanning the directive.
	    \return A \c line_type determined by evaluation.
	*/
	static line_type eval(directive_type type, char const **ptext);

	/*! \brief Evaluate an \c \#ifdef or \c \#ifndef directive
	    \param  type    The \c directive_type of the directive.
	    \param  ptext On entry, a pointer to starting address of the directive's
	        body. On return, receives the first address not consumed by
	        scanning the directive.
	    \return A \c line_type determined by evaluation.
	*/
	static line_type
	eval_ifdef_or_ifndef(directive_type type, char const **ptext);

private:
	//! Type of evaluation functions applied to directive bodies
	typedef line_type (*evaluator)(char const **);
	//! Get the \c directive_type of a directive given a keyword
	static directive_type get_type(std::string const & keyword);
	//! Map from keywords to directive types.
	static std::map<std::string, directive_type> _keyword_to_type_map_;
	/*! \brief Table of evaluation functions for directive types, indexed
	    by \c directive_type.
	*/
	static std::vector<evaluator> _evaluator_tab_;
#if !CPP11_HAVE_UNIFORM_INITIALIZATION
	typedef
		std::map<std::string, directive_type>::value_type keyword_map_element;
	static size_t _n_evaluators_;
	static size_t _n_keywords_;
	static evaluator _evaluator_tab_initor_[];
	static keyword_map_element _keyword_to_type_map_initor_[];

#endif
};

/*! \brief Encapsulates a directive of a given type

    Class template \e directive<Type> encapsulates a directive of
    a given type and provides a global lookup container for tracking
    occurrences of directives of that type.

    The class template supports the coan commands that report occurrences of
    directives. The	container is keyed by the canonicalised text of directive
	occurrence. Storing the directive occurrences in canonicalised form enables
	occurrences to be compared regardless of insignificant differences of
	whitespace or comments.

    \tparam Type    The \c directive_type of directives contained by this class.
*/
template<directive_type Type>
struct directive : directive_base {
	friend struct directive_base;
	//! The \c directive_type of this class.
	static directive_type const _type_ = Type;

#if CPP11_HAVE_DECL_DELETE
	//! Default construction is not allowed
	directive() = delete;
	//! Copying a is not allowed
	directive(directive const &) = delete;
	//! Assignment is not allowed
	directive & operator=(directive const &) = delete;
#else
	directive();
	directive(directive const &);
	directive & operator=(directive const &);
#endif
	//! Destructor
	virtual ~directive() {};

	/*! \brief Explicitly construct a directive from a directive body.
	    \param arg  The directive body.
	*/
	explicit directive(std::string const & arg)
		:  _loc(insert(arg)) {}

	//! Get the body of the directive.
	std::string const & argument() const {
		return _loc->first;
	}

	//! Report the directive.
	virtual void report() {
		directive_base::report(_loc->second,_keyword_,_loc->first);
		_loc->second = true;

	}

	//! Say whether the directive has been reported.
	bool reported() const {
		return _loc->second;
	}

	//! Set the status of the directive as reported or unreported.
	void set_reported(bool reported = true) {
		_loc->second = reported;
	}

protected:

	/*! \brief Evaluate a directive of this type.
	    \param ptext On entry, a pointer to starting address of the
	        directive's body. On return, receives the first address not consumed
	        by scanning the directive.
	    \return A \c line_type determined by evaluation
	            denoting the action to be taken on this line.
	*/
	static line_type eval(char const **ptext);

	/*! \brief Type of global lookup table for directives of this type.

	    The container maps the canonicalized text of the directive to
	    a \c bool indicating whether the directive has been reported.
	*/
	typedef std::map<std::string,bool> directives_table;
	//! Type of entry in \c directives_table.
	typedef directives_table::value_type table_entry;

	/*! \brief Insert a directive into the global lookup table.

	    \param  arg The body of the directive.
	    \return An interator to the inserted entry.
	*/
	static directives_table::iterator insert(std::string const & arg) {
		return _directives_tab_.insert(table_entry(arg,false)).first;
	}

	//! An iterator locating this directive in the global lookup table.
	directives_table::iterator _loc;

	//! The keyword for directives of the type.
	static std::string const _keyword_;
	//! The global lookup table for directives of this type.
	static directives_table _directives_tab_;
};

//! \cond NO_DOXYGEN

template<>
line_type directive<HASH_UNKNOWN>::eval(char const **ptext);

template<>
line_type directive<HASH_IF>::eval(char const **ptext);

template<>
line_type directive<HASH_IFDEF>::eval(char const **ptext);

template<>
line_type directive<HASH_IFNDEF>::eval(char const **ptext);

template<>
line_type directive<HASH_ELSE>::eval(char const **ptext);

template<>
line_type directive<HASH_ELIF>::eval(char const **ptext);

template<>
line_type directive<HASH_ENDIF>::eval(char const **ptext);

template<>
line_type directive<HASH_DEFINE>::eval(char const **ptext);

template<>
line_type directive<HASH_UNDEF>::eval(char const **ptext);

template<>
line_type directive<HASH_INCLUDE>::eval(char const **ptext);

template<>
line_type directive<HASH_PRAGMA>::eval(char const **ptext);

template<>
line_type directive<HASH_ERROR>::eval(char const **ptext);

//! \endcond

#endif /* EOF*/
