#ifndef EVALUATION_H
#define EVALUATION_H

/***************************************************************************
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Originally contributed by Mike Kinghan, imk@burroingroingjoing.com.   *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "platform.h"
#include "integer.h"
#include <string>

/*! \file evaluation.h
   This file defines class \c evaluation.
*/

/*! \brief Class \e evaluation represents the result of evaluating a putative
    expression.
*/
struct evaluation {
	/*! Default constructor */
	evaluation()
		: _value(INT_UNDEF),_constant(false),
		  _well_formed(true),_circular(false),_del_paren(false),_empty(false) {}

	/*! \brief Explicitly construct from \c integer, optionally a constant.
	    \param  val An integer to be evaluated.
	    \param  constant    True iff \c val is a constant.
	*/
	explicit evaluation(integer const & val, bool constant = false)
		: _value(val),_constant(constant),
		  _well_formed(true),_circular(false),_del_paren(false),_empty(false) {}

	/*! \brief Explicitly construct from \c int.
	    \param  val An integer to be evaluated.
	*/
	explicit evaluation(int val)
		: _value(integer(INT_INT,val)),_constant(false),
		  _well_formed(true),_circular(false),_del_paren(false),_empty(false) {}

	/*! Destructor */
	~evaluation() {}

	//! Say whether the expression has been resolved.
	bool resolved() const {
		return _value.type() != INT_UNDEF;
	}

	//! Say whether the expression is well-formed.
	bool well_formed() const {
		return resolved() && _well_formed;
	}

	//! Say whether the expression is malformed.
	bool malformed() const {
		return resolved() && !_well_formed;;
	}

	//! Say whether the expression is true.
	bool is_true() const {
		return well_formed() && _value.raw() != 0;
	}

	//! Say whether the expression is false.
	bool is_false() const {
		return well_formed() && _value.raw() == 0;
	}

	//! Say whether the expression is an integral constant.
	bool constant() const {
		return well_formed() && _constant;
	}

	//! Say whether surrounding parentheses may be deleted.
	bool del_paren() const {
		return _del_paren;
	}

	//! Say whether the expression involves circular definitions.
	bool circular() const {
		return _circular;
	}

	//! Say whether the expression is empty.
	bool empty() const {
		return _empty;
	}

	//! Get the integral value of the expression.
	integer const & value() const {
		return _value;
	}

	//! Set the integral value of the expression.
	void set_value(integer const & val) {
		_value = val;
	}

	//! Set the integral value of the expression to an integer.
	void set_value(int val) {
		_value = integer(INT_INT,val);
	}

	/*! \brief Classify the expression as a constant or non-constant.
	    \param value    If true the expression is classified as a constant,
	        else as non-constant.
	*/
	void set_constant(bool value = true) {
		_constant = value;
	}

	//! Classify the expression as insoluble.
	void set_insoluble() {
		_value = integer(INT_UNDEF);
	}

	/*! \brief Classify the expression as eligible for deletion of surrounding
	    parentheses or not.
	    \param  value If true the expression is eligible for deletion of
	        srrounding parentheses, else not.
	*/
	void set_del_paren(bool value = true) {
		_del_paren = value;
	}

	//! Classify the expression as malformed.
	void set_malformed() {
		set_insoluble();
		_well_formed = false;
	}

	//! Classify the expression as involving one or more circular definitions.
	void set_circular() {
		set_insoluble();
		_circular = true;
	}

	//! Classify the expression as empty.
	void set_empty() {
		set_insoluble();
		_empty = true;
	}

private:
	//! The integer value of the evaluated expression if it is soluble
	integer _value;
	//! Is the expression an integer constant?
	bool _constant;
	//! Is the expression well-formed
	bool _well_formed;
	//! Does the expression involve circular definitions?
	bool _circular;
	//! Can parentheses surrounding the expression be deleted
	bool _del_paren;
	//! Is the expression empty?
	bool _empty;
};

/*@}*/

#endif /* EOF*/
