/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2002, 2003 Tony Finch <dot@dotat.at>.                   *
 *   All rights reserved.                                                  *
 *   Copyright (C) 1985, 1993 The Regents of the University of California. *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed by Mike Kinghan, imk@burroingroingjoing.com,              *
 *   derived from the code of Tony Finch                                   *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "expression_parser.h"
#include "symbol.h"
#include "io.h"
#include "line_edit.h"
#include "chew.h"
#include "options.h"
#include "line_select.h"
#include "lexicon.h"
#include "integer_constant.h"
#include "canonical_string.h"
#include <string>
#include <cassert>

using namespace std;

/*!
    \file expression_parser.cpp
    This file implements class \e expression_parser
*/

evaluation
expression_parser::op_or(evaluation const & lhs, evaluation const & rhs)
{
	if (lhs.is_true() || rhs.is_true()) {
		return evaluation(1);
	} else if (lhs.is_false() && rhs.is_false()) {
		return evaluation(0);
	}
	return evaluation();
}

evaluation
expression_parser::op_and(evaluation const & lhs, evaluation const & rhs)
{
	if (rhs.is_true() && lhs.is_true()) {
		return evaluation(1);
	} else if (rhs.is_false() || lhs.is_false()) {
		return evaluation(0);
	}
	return evaluation();
}

evaluation
expression_parser::op_divide(evaluation const & lhs, evaluation const & rhs)
{
	if (rhs.value().good() && rhs.value().raw() == 0) {
		warning_zero_divide() << "Divide by zero" << emit();
		return evaluation();
	}
	return evaluation(lhs.value() / rhs.value());
}

evaluation
expression_parser::op_mod(evaluation const & lhs, evaluation const & rhs)
{
	if (rhs.value().good() && rhs.value().raw() == 0) {
		warning_zero_divide() << "Divide by zero" << emit();
		return evaluation();
	}
	return evaluation(lhs.value() % rhs.value());
}

const expression_parser::co_precedent_ops expression_parser::eval_ops[] = {

	{
		&expression_parser::binary_op,
		{
			{ TOK_TERNARY_IF, op_ternary_if_dummy }
		}
	},
	{
		&expression_parser::binary_op,
		{
			{ TOK_TERNARY_EITHER_OR, op_ternary_either_or_dummy }
		}
	},
	{
		&expression_parser::binary_op,
		{
			{ TOK_BOOLEAN_OR, op_or }
		}
	},
	{
		&expression_parser::binary_op,
		{
			{ TOK_BOOLEAN_AND, op_and }
		}
	},
	{
		&expression_parser::binary_op,
		{
			{ TOK_BIT_OR, op_bit_or },
			{ TOK_BIT_XOR, op_bit_xor },
			{ TOK_BIT_AND, op_bit_and }
		}
	},
	{
		&expression_parser::binary_op,
		{
			{ TOK_EQ, op_eq },
			{ TOK_NEQ, op_ne }
		}
	},
	{
		&expression_parser::binary_op,
		{
			{ TOK_LE, op_le },
			{ TOK_GE, op_ge },
			{ TOK_LESS, op_lt },
			{ TOK_GREATER, op_gt }
		}
	},
	{
		&expression_parser::binary_op,
		{
			{ TOK_LSHIFT, op_lshift },
			{ TOK_RSHIFT, op_rshift },
		}
	},
	{
		&expression_parser::binary_op,
		{
			{ TOK_ADD, op_add },
			{ TOK_SUBTRACT, op_subtract },
		}
	},
	{
		&expression_parser::unary_op,
		{
			{ TOK_MULTIPLY, op_mult },
			{ TOK_DIV, op_divide },
			{ TOK_MOD, op_mod },
		}
	},
};

expression_parser::co_precedent_ops const *
expression_parser::ops_tab = expression_parser::eval_ops;

bool expression_parser::match_op(char const ** ptxt, char const *op)
{
	char const *txt = *ptxt;
	bool verdict = true;
	assert(op[1] == '\0' || op[2] =='\0');
	if (txt[0] != op[0]) {
		verdict = false;
	} else {
		if (op[1]) {
			txt = chewer<continuation>(++txt);
			if (txt[0] != op[1]) {
				verdict = false;
			}
		} else if (txt[1] == op[0]) {
			verdict = false;
		}
	}
	if (verdict) {
		*ptxt = txt + 1;
	}
	return verdict;
}

evaluation expression_parser::apply(const operation *op,
									evaluation & lhs, evaluation &rhs)
{
	if (op == eval_ops[0].op) {
		return rhs;
	}
	if (op == eval_ops[1].op) {
		if (!_ternary_cond_stack.size()) {
			if (_in_code) {
				error_ternary_cond_incomplete() << 
					"':' of ternary conditional without preceding '?' "
					"in \"" << io::pristine_line() << "\"" << emit();
			}
			return evaluation();
		}
		else {
			evaluation ternary_cond = _ternary_cond_stack.back();
			_ternary_cond_stack.pop_back();
			if (ternary_cond.is_true()) {
				return lhs;
			}
			if (ternary_cond.is_false()) {
				return rhs;
			}
			return evaluation();
		}
	}
	return op->fn(lhs,rhs);
}

evaluation expression_parser::binary_op(co_precedent_ops const *ops,
                                        char const **cpp,
                                        operation const **pbin_op)
{
	size_t start_cut = io::read_offset(*cpp);
	/* Evaluate the lhs...*/
	operation const *lhs_bin_op = NULL;
	operation const *rhs_bin_op = NULL;
	evaluation lhs = (this->* ops->inner)(ops+1,cpp,&lhs_bin_op);
	/* Assume lhs is all we've got...*/
	evaluation result = lhs;
	char const * cp = *cpp;
	size_t start_lhs_cut = start_cut;
	size_t end_lhs_cut = io::read_offset(cp);
	if (!result.malformed()) {
		for (; *(cp = chewer<discardable>(cp)) && *cp != ')'; ) {
			const operation *op;
			evaluation rhs;
			evaluation ternary_cond;
			end_lhs_cut = io::read_offset(cp);
			/* Now look for binary op at current precedence...*/
			for (op = ops->op;
			     op->str != NULL && !match_op(&cp,op->str); op++) {}
			if (op->str == NULL) {
				/* No binary op at current precedence, skip rhs. */
				if (ops == eval_ops) {
					/* No binary op at all. Illegal */
					result.set_malformed();
				}
				break;
			}
			/* Got bin op - store it for return */
			*pbin_op = ops->op;
			if (op == eval_ops[0].op) {
				_ternary_cond_stack.push_back(lhs);
			}
			/* Assume we will delete parentheses...*/
			result.set_del_paren();
			start_cut = end_lhs_cut;
			end_lhs_cut = io::read_offset(cp);
			/* Evaluate rhs...*/
			rhs = (this->*ops->inner)(ops,&cp,&rhs_bin_op);
			rhs.set_constant(false);
			lhs.set_constant(false);
			result = apply(op,lhs,rhs);
			if (!rhs.resolved() && !lhs.resolved()) {
				if (!lhs.del_paren() && rhs.del_paren() &&
				    (op != rhs_bin_op || op != lhs_bin_op)) {
					line_edit::restore_paren(start_cut,io::read_offset(cp));
				} else if (lhs.del_paren() && !rhs.del_paren() &&
				           (op != rhs_bin_op || op != lhs_bin_op)) {
					line_edit::restore_paren(start_lhs_cut,end_lhs_cut);
				}
			} else {
				if (lhs.resolved()) {
					/* Cutting either operand makes parentheses superfluous*/
					result.set_del_paren();
					if (!rhs.resolved() &&
					    (op->fn == op_or || op->fn == op_and)) {
						line_edit::cut_text(start_lhs_cut,end_lhs_cut);
					}
				}
				if (rhs.resolved()) {
					result.set_del_paren();
					if (!lhs.resolved() &&
					    (op->fn == op_or || op->fn == op_and)) {
						line_edit::cut_text(start_cut,io::read_offset(cp));
					}
				}
			}
		}
		if (*pbin_op && result.del_paren()) {
			line_edit::delete_paren(start_lhs_cut,end_lhs_cut);
		}
	}
	*cpp = cp;
	return result;
}


evaluation expression_parser::unary_op(co_precedent_ops const *ops,
                                       char const **cpp,
                                       operation const **ignored)
{
	char const *ep;
	evaluation result;
	char const *cp;
	result.set_del_paren();
	chewer_base::set_stop_at_quote(true);
	cp = chewer<discardable>(*cpp);
	chewer_base::set_stop_at_quote(false);
	do {
		if (*cp == '!') {
			cp = chewer<continuation>(++cp);
			result = unary_op(ops,&cp,NULL);
			if (!result.resolved()) {
				break;
			}
			result.set_value(!result.value());
			result.set_constant(false);
			break;
		}
		if (*cp == '~') {
			cp = chewer<continuation>(++cp);
			result = unary_op(ops,&cp,NULL);
			if (!result.resolved()) {
				break;
			}
			result.set_value(~result.value());
			result.set_constant(false);
			break;
		}
		if (*cp == '(') {
			char const *start = cp;
			operation const *bin_op = NULL;
			cp = chewer<continuation>(++cp);
			result = binary_op(eval_ops,&cp,&bin_op);
			cp = chewer<discardable>(cp);
			if (*cp != ')') {
				/* Missing ')'*/
				if (_in_code) {
					error_unbalanced_paren() << "Missing \")\" in \""
					                         << io::pristine_line()
					                         << '\"' << emit();
				} else {
					result.set_malformed();
					break;
				}
			} else if (result.del_paren()) {
				line_edit::delete_paren(io::read_offset(start),
				                        io::read_offset(cp));
			}
			cp = chewer<continuation>(++cp);
			break;
		}
		if (*cp == '+') {
			cp = chewer<continuation>(++cp);
			result = unary_op(ops,&cp,NULL);
			break;
		}
		if (*cp == '-') {
			cp = chewer<continuation>(++cp);
			result = unary_op(ops,&cp,NULL);
			result.set_value(-result.value());
			break;
		}
		if (isdigit((unsigned char)*cp)) {
			integer val = integer_constant::read_numeral(cp,&ep);
			if (val.type() != INT_UNDEF) {
				result.set_value(val);
				if (_in_code) {
					if (!options::eval_constants()) {
						result.set_constant();
					}
				}
			}
			cp = (char *)ep;
			break;
		}
		string word = canonical<symbol>(cp,&ep);
		if (word == "defined") {
			bool paren;
			cp = chewer<discardable>(ep);
			paren = *cp == '(';
			if (paren) {
				cp = chewer<continuation>(++cp);
			}
			cp = chewer<discardable>(cp);
			symbol sym(symbol::read(cp,&cp));
			cp = chewer<discardable>(cp);
			if (paren) {
				if (*cp == ')') {
					cp = chewer<continuation>(++cp);
				}
				/* Missing ')'*/
				else if (_in_code) {
					error_unbalanced_paren() <<
					                         "Missing \")\" in \"" <<
					                         io::pristine_line() << '\"'
					                         << emit();
				} else {
					result.set_malformed();
					break;
				}
			}
			if (!sym.configured()) {
				if (line_select::eligible(CMD_SYMBOLS)) {
					sym.record(UNCONFIGURED);
					sym.report();
				}
				if (options::implicit()) {
					result.set_value(0);
				}
				break;
			} else if (line_select::eligible(CMD_SYMBOLS)) {
				sym.report();
			}
			result.set_value(sym.defined());
			break;
		}
		if (*cp == '\'') {
			/* Possible character constant */
			integer val = integer_constant::read_char(cp,&cp);
			if (val.type() != INT_UNDEF) { /* Got a character constant */
				result.set_value(val);
				break;
			}
		} else if (symbol::is_start_char(*cp)) {
			if (*cp == 'L') {
				/* Possible wide character constant */
				integer val = integer_constant::read_char(cp,&cp);
				if (val.type() != INT_UNDEF) { /* Got a character constant */
					result.set_value(val);
					break;
				}
			}
			if (symbol::is_start_char(*cp)) {
				string symname = symbol::read(cp,&cp);
				symbol sym = symbol::lookup(symname);
				argument_list macro_args(cp,&cp);
				if (!macro_args.well_formed()) {
					error_malformed_macro() <<
						"Malformed macro argument for symbol \""
						<< symname << "\"" << emit();
					break;
				}
				if (!sym.configured()) {
					if (_in_code) {
						if (sym.null()) {
							sym = symbol(symname);
						}
						if (line_select::eligible(CMD_SYMBOLS)) {
							sym.report(macro_args);
						}
						if (options::implicit()) {
							result.set_value(0);
						}
					}
				} else {
					if (macro_args.length() != sym.macro_parameters().length()) {
						error_mismatched_macro_args()
							<< "Symbol " << sym.id() << " does not take "
							<< macro_args.length() << " macro parameters" << emit();
						break;
					}
					result = sym.evaluate(macro_args);
					if (line_select::eligible(CMD_SYMBOLS) && _in_code) {
						sym.report();
					}
					if (sym.defined() && sym.defn()->length() == 0) {
						warning_empty_symbol() << "Empty symbol \""
							   << sym.id() << "\" in expression" << emit();
					}
				}
				break;
			}
		} else if (_in_code) {
			error_garbage_in_directive() << "Garbage in \""
				 << io::pristine_line()
				 << '\"' << emit();
		}
		result.set_malformed();
	} while(false);
	*cpp = cp;
	return result;
}

expression_parser::expression_parser(char const **ptext, bool in_code)
	: _in_code(in_code),_ternary_cond_stack(0)
{
	operation const * bin_op = NULL;
	_eval = binary_op(eval_ops,ptext,&bin_op);
	if (!_in_code) {
		/* Forget any unclosed quotation if not parsing code */
		chewer_base::set_in_double_quote(false);
	}
	if (_ternary_cond_stack.size()) {
		_eval.set_malformed();
		if (_in_code) {
			error_ternary_cond_incomplete() << 
				"'?' of ternary conditional without following ':' "
				"in \"" << io::pristine_line() << "\"" << emit();
		}
	}
}

/* EOF*/
