/***************************************************************************
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "integer_constant.h"
#include "diagnostic.h"
#include "chew.h"
#include <climits>
#include <cassert>

/*! \file integer_constant.cpp
    This file defines class \c integer_constant.
*/

using namespace std;

unsigned long long
integer_constant::decode_utf8(char const *cp, char const **pend)
{
	unsigned long long val = ULLONG_MAX;
	size_t bytes = 0;
	unsigned byte1 = *cp & 0xff;
	unsigned byte2, byte3, byte4;
	if (byte1 == unsigned(EOF)) {
		bytes = 0;
	} else if (byte1 >> 7 == 0) {
		/* Top bit of lead byte is 0. This is a plain ascii code */
		bytes = 1;
	} else if (byte1 >> 5 == 0x6) {
		/* Top 3 bits of lead byte are 110. A 2-byte code. */
		bytes = 2;
	} else if (byte1 >> 4 == 0xe) {
		/* Top 4 bits of lead byte are 1110. A 3-byte code. */
		bytes = 3;
	} else if (bytes >> 3 == 0x1e) {
		/* Top 5 bits of lead byte are 11110. A 4-byte code. */
		bytes = 4;
	}
	switch(bytes) {
	case 1:
		val = byte1;
		cp = chewer<continuation>(++cp);
		break;
	case 2:
		byte1 &= 0x1F; /* Clear top three bits of lead byte. */
		byte2 = 0xff & *(cp = chewer<continuation>(++cp));
		if (byte2 == unsigned(EOF) || byte2 >> 6 != 2) {
			/* Top 2 bits of continuation byte must be 10 */
			break;
		}
		byte2 &= 0x3F;
		/* Value is concatenation of non-control bits */
		val = ((byte1 << 6) | byte2);
		cp = chewer<continuation>(++cp);
		break;
	case 3:
		byte1 &= 0x1F;
		byte2 = 0xff & *(cp = chewer<continuation>(++cp));
		if (byte2 == unsigned(EOF) || byte2 >> 6 != 2) {
			break;
		}
		byte2 &= 0x3F;
		byte3 = 0xff & *(cp = chewer<continuation>(++cp));
		if (byte3 == unsigned(EOF) || byte3 >> 6 != 2) {
			break;
		}
		byte3 &= 0x3F;
		val = (byte1 << 12) | (byte2 << 6) | byte3;
		cp = chewer<continuation>(++cp);
		break;
	case 4:
		byte1 &= 0x1F;
		byte2 = 0xff & *(cp = chewer<continuation>(++cp));
		if (byte2 == unsigned(EOF) || byte2 >> 6 != 2) {
			break;
		}
		byte2 &= 0x3F;
		byte3 = 0xff & *(cp = chewer<continuation>(++cp));
		if (byte3 == unsigned(EOF) || byte3 >> 6 != 2) {
			break;
		}
		byte3 &= 0x3F;
		byte4 = 0xff & *(cp = chewer<continuation>(++cp));
		if (byte4 == unsigned(EOF) || byte4 >> 6 != 2) {
			break;
		}
		byte4 &= 0x3F;
		val = (byte1 << 18) | (byte2 << 12) | (byte3 << 6) | byte4;
		cp = chewer<continuation>(++cp);
		break;
	default:
		break;
	}
	*pend = cp;
	return val;
}


pair<unsigned long long,bool>
integer_constant::read_based_numeral(unsigned base,char const *cp, char const **pend)
{
	int dval = 0;
	unsigned long long val = 0;
	unsigned long long tmp;
	bool overflow = false;
	for (	;; cp = chewer<continuation>(++cp)) {
		switch(*cp) {
		case '0':
			dval = 0;
			break;
		case '1':
			dval = 1;
			break;
		case '2':
			dval = 2;
			break;
		case '3':
			dval = 3;
			break;
		case '4':
			dval = 4;
			break;
		case '5':
			dval = 5;
			break;
		case '6':
			dval = 6;
			break;
		case '7':
			dval = 7;
			break;
		case '8':
			dval = 8;
			break;
		case '9':
			dval = 9;
			break;
		case 'a':
		case 'A':
			dval = 10;
			break;
		case 'b':
		case 'B':
			dval = 11;
			break;
		case 'c':
		case 'C':
			dval = 12;
			break;
		case 'd':
		case 'D':
			dval = 13;
			break;
		case 'e':
		case 'E':
			dval = 14;
			break;
		case 'f':
		case 'F':
			dval = 15;
			break;
		default:
			dval = 16;
		}
		if (unsigned(dval) < base) {
			tmp = val;
			val *= base;
			if (val / base != tmp) {
				overflow = true;
			} else if (dval) {
				tmp = val;
				val += dval;
				if (val <= tmp) {
					overflow = true;
				}
			}
		} else {
			break;
		}
	}
	if (pend) {
		*pend = cp;
	}
	return pair<unsigned long long, bool>(val,overflow);
}

int
integer_constant::read_char_escaping(char const *cp, char const **pend)
{
	int val = EOF;
	if (*cp == '\\') {
		cp = chewer<continuation>(++cp);
		switch(*cp) {
		case '\0':
			break;
		case 'a':
			val = '\a';
			break;
		case 'b':
			val = '\b';
			break;
		case 'f':
			val = '\f';
			break;
		case 'n':
			val = '\n';
			break;
		case 'r':
			val = '\r';
			break;
		case 't':
			val = '\t';
			break;
		case 'v':
			val = '\v';
			break;
		default:
			val = *cp;
		}
		cp = chewer<continuation>(++cp);
	} else if (*cp) {
		val = *cp;
		cp = chewer<continuation>(++cp);
	}
	*pend = cp;
	return val;
}


unsigned long long
integer_constant::read_encoded_char(int base, char const *cp, char const **pend)
{
	unsigned long long val = ULLONG_MAX;
	char prefix = 0;
	assert(base == 8 || base == 16);
	if (base == 8) {
		prefix = '0';
	} else if (base == 16) {
		prefix = 'x';
	} else {
		assert(false);
	}
	*pend = cp;
	if (*cp == '\'') {
		cp = chewer<continuation>(++cp);
		if (*cp == '\\' && *(cp = chewer<continuation>(++cp)) == prefix) {
			cp = chewer<continuation>(++cp);
			std::pair<unsigned long long,bool> verdict =
			    read_based_numeral(base,cp,&cp);
			val = verdict.first;
			bool overflow = verdict.second;
			if (overflow || val > INT_MAX || *cp != '\'') {
				val = ULLONG_MAX;
			} else {
				*pend = chewer<continuation>(++cp);
			}
		}
	}
	return val;
}

integer
integer_constant::read_numeral(char const *cp, char const **pend)
{
	/*  Initially assume numeral may represent a long long
	    until a type suffix confirms or disconfirms.
	    Finally, in the absence of any type suffix, we will say the numeral
	    is an int (signed or unsigned) if it is not too big
	*/
	integer result(INT_LLONG,ULLONG_MAX);
	unsigned base = 10;
	unsigned long long max_val = ULLONG_MAX;
	bool overflow = false;
	size_t num_len = 0;
	char const *start = cp;
	char const *type_desc = NULL;
	bool has_suffix = false;
	if (*cp == 0) {
		return result;
	}
	if (*cp == '0') {
		cp = chewer<continuation>(++cp);
		if (*cp == 'x' || *cp == 'X') {
			cp = chewer<continuation>(++cp);
			base = 16;
		} else {
			base = 8;
		}
	}
	pair<unsigned long long, bool> verdict = read_based_numeral(base,cp,&cp);
	result._val = verdict.first;
	overflow = verdict.second;
	if (overflow) {
		result._type = INT_ULLONG;
		type_desc = "unsigned long long";
		max_val = ULLONG_MAX;
	}
	num_len = cp - start;
	if (num_len == 1 && (*cp == 'x' || *cp == 'X')) {
		cp = start;
	} else if (num_len > 0) {
		if (*cp == 'u' || *cp == 'U') {
			has_suffix = true;
			result._type = INT_UINT;
			cp = chewer<continuation>(++cp);
		}
		if (*cp == 'l' || *cp == 'L') {
			has_suffix = true;
			cp = chewer<continuation>(++cp);
			result._type = result.is_signed() ? INT_LONG : INT_ULONG;
		}
		if (*cp == 'l' || *cp == 'L') {
			has_suffix = true;
			cp = chewer<continuation>(++cp);
			result._type = result.is_signed() ? INT_LLONG : INT_ULLONG;
		}
		if (result.is_signed() && (*cp == 'u' || *cp == 'U')) {
			if (!has_suffix) {
				has_suffix = true;
				result._type = INT_UINT;
			} else {
				result.make_unsigned();
			}
			cp = chewer<continuation>(++cp);
		}
	}
	if (pend) {
		*pend = cp;
	}
	if (num_len == 0) {
		return result;
	}
	if (result.is_signed() && result._val > LLONG_MAX && !has_suffix) {
		/* If we have no suffix explicitly specifying long or long long
		    then we allow the value to be an unsigned long long and emit
		    a warning. We store the warning at this point and will write it
		    later if it is not superceded by an overflow warning */
		warning_forced_unsigned() <<
			  "Integer constant \"" <<
			  string(start,cp) << "\" is so large it is treated as unsigned"
			  << defer();

		result._type = INT_ULLONG;
	}
	if (result.type() == INT_ULONG && result._val > ULONG_MAX) {
		overflow = true;
		type_desc = "unsigned long";
		max_val = ULONG_MAX;
	} else if (result.type() == INT_LONG && result._val > LONG_MAX) {
		overflow = true;
		type_desc = "long";
		max_val = LONG_MAX;
	} else if (result.type() == INT_LLONG && result._val > LLONG_MAX) {
		overflow = true;
		type_desc = "long long";
		max_val = LLONG_MAX;
	}
	if (overflow) {
		/* On overflow discard any stored warning */
		warning_forced_unsigned::discard();
		warning_int_overflow() << "Integer constant \"" <<
			   string(start,cp) << "\" is too large for type " <<
			   type_desc << "(max " << max_val <<
			   "): expression will not be resolved" << emit();
		result._type = INT_UNDEF;
	} else {
		warning_forced_unsigned::flush(); /* Emit any stored warning */
		if (!has_suffix) {
			/* No type-suffix. Reduce type to (unsigned) int if value will fit
			*/
			if (result.is_signed()) {
				if (result._val <= (unsigned long long)INT_MAX) {
					result._type = INT_INT;
				}
			} else if (result._val <= (unsigned long long)UINT_MAX) {
				result._type = INT_UINT;
			}
		}
	}
	return result;
}

integer integer_constant::read_char(char const *cp, char const **pend)
{
	integer result(INT_UNDEF);
	char const *type_desc = "int";
	unsigned long long val = ULLONG_MAX;
	bool wide = false;
	int max_len = sizeof(int);
	*pend = cp;
	/* Assume multi-byte constant. Can have sizeof(int) bytes */
	if (*cp == 'L') { /* No, this is a wide-character constant */
		cp = chewer<continuation>(++cp);
		max_len = sizeof(wchar_t); /* Can have sizeof(wchar_t) bytes */
		type_desc = "wchar_t";
		wide = true;
	}
	if (*cp == '\'') {
		val = read_encoded_char(16,cp,&cp); // hex?
		if (val == ULLONG_MAX) {
			val = read_encoded_char(8,cp,&cp); // octal?
		}
		if (val == ULLONG_MAX) {
			/* Not a hex or octal constant. Try multi-byte or wide constant */
			int nbytes = 0;
			val = 0;
			cp = chewer<continuation>(++cp); /* Consume opening quote */
			/* If not a wide character constant then we'll evaluate as a
			    concatenation of 8-bit ints. But likewise even if it is a wide
			    character constant that commences with an escape, because an
			    escaped character must be ascii. */
			if (!wide || (cp[0] == '\\' && cp[1] != '\n')) {
				int chval = 0;
				for (; nbytes < max_len && *cp && *cp != '\'' &&
				     (chval = integer_constant::read_char_escaping(cp,&cp)) !=
						EOF;
				     val = (val << 8 | chval), ++nbytes) {}
			} else {
				val = decode_utf8(cp,&cp);
			}
			if (val == ULLONG_MAX) {}
			if (*cp != '\'') {
				/* Not a valid constant. No closing quote within max_len. */
				val = ULLONG_MAX;
				/* Look for that closing quote... */
				for (; *(cp = chewer<continuation>(++cp)) != '\0' &&
				     *cp != '\n' && *cp != '\'';
				     cp = chewer<continuation>(++cp)) {};
				if (!*cp || *cp == '\n') { /* Unclosed quotation */
					warning_unclosed_quote() << "Unclosed quotation after \"" <<
						 string(*pend,cp) << '\"' << emit();
				} else { /* Constant is too long for type */
					warning_char_constant_too_long() << "Character constant " <<
						 string(*pend,cp + 1) << " is too long for type " <<
						 type_desc << "(max " << max_len <<
						 "). Will not be resolved" << emit();
				}
			} else if (!wide && nbytes > 1) {
				warning_mulitbyte_char_constant() <<
					  "Multi-byte character constant " <<
					  string(*pend,cp + 1) << emit();
			}
			cp = chewer<continuation>(++cp);
		}
		*pend = cp;
	}
	if (val != ULLONG_MAX) {
		result._val = val;
		result._type = INT_INT;
	} else {
		result._type = INT_UNDEF;
	}
	//IMK *pend = cp;
	return result;
}

// EOF
