#ifndef INTEGER_CONSTANT_H
#define INTEGER_CONSTANT_H
/***************************************************************************
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "integer.h"

/*! \file integer_constant.h
    This file defines class \c integer_constant.
*/

//! Class \c integer_constant encapsulates an integer constant
struct integer_constant {
	/*! \brief Read a numeral from a text pointer returning an \c integer.
	    \param	cp	Address of the putative numeral.
	    \param	pend If not NULL, receives on return the address just past the
	        parsed text.

	    \return An \c integer that represents the integer value of the numeral,
	    unless the value is too big for the explicit or implied type of the
	    numeral in which case \c integer(INT_UNDEF) is returned.

	    The parsed numeral is attributed type signed int unless an
	    a type suffix indicates otherwise.

	    The numeral may include any of the type-indicating suffixes
	    'L', 'LL' 'UL', 'LU', 'ULL', 'LLU' or their case-variants.

	    The string "0x" or "0X" is parsed as a valid numeral "0" followed
	    by 'x' or 'X'.

	*/
	static integer read_numeral(char const *cp, char const **pend = NULL);

	/*! \brief Read a character constant from a text pointer, returning
	    its value as an \c integer.

	    \param  cp  Address of the putative character constant.
	    \param  pend If not NULL, receives on return the address just past
	            the parsed constant.
	    \return An \c integer that represents the integer value of
	        the parsed character constant, or \c integer(UNDEF)
	        is no valid constant is parsed.
	*/
	static integer read_char(char const *cp, char const **pend = NULL);

private:

	/*! \brief Decode an UTF-8 encoded character from a buffer.

	    Decode an UTF-8 encoded character from a buffer ignoring
	    embedded line-continautions.
	    \param  cp Address of the first byte to to be decoded.

	    \param pend Receives on return the address of the first byte
	        not decoded.
	    \return The integer value of the UTF-8 character parsed, or \c
	        ULLONG_MAX if none is parsed.
	*/
	static unsigned long long
	decode_utf8(char const *cp, char const **pend);

	/*! \brief Read a possibly escaped ASCII character from a text pointer
	    returning its escaped value.
	    \param  cp  Address from which to read a character.

	    \param pend On return, receives the address of the next unconsumed
	        character. A null character will not be consumed.
	    \return The integer value of the possibly escaped character consumed,
	            if any, else EOF.

	    If \e cp addresses the escape character '\\' on entry then that
	    character is consumed and, if the next character is non-null, its
	    escaped value is returned: if the next character is null then EOF is
	    returned.

	    If \e cp does not address the escape character on entry then, if \e *cp
	    is non-null, its integer value is returned: if \e *cp == '\0'
	    then EOF is returned.
	*/
	static int
	read_char_escaping(char const *cp, char const **pend);


	/*! \brief Read a numeral of known base from a text pointer, returning a
	    value and overflow indicator.
	    \param  base The arithmetic base in which the numeral is to be evaluated.

	    \param  cp  Address of the putative numeral.
	    \param  pend If not \c NULL, receives on return the address just past
	            the parsed numeral.
	    \return An \c std::pair containing, first, the integer value of the
	        parsed numeral; second, a bool that is true if the value of
	        the parsed numeral overflows unsigned long long, in which case
	        the returned integer value is not valid.

	    The function consumes characters after \b cp for as long as
	    they can form part of a numeral with the indicated base.
	*/
	static std::pair<unsigned long long,bool>
	read_based_numeral(unsigned base,char const *cp, char const **pend);

	/*! \brief Read a numerically encoded character constant from a buffer,
	    returning its value as an \c integer.

	    \param base The presumed base of the constant, either 8 (octal) or
	        16 (hex).

	    \param  cp  Pointer to the address of the putative encoded constant.

	    \param  pend If not \c NULL, receives on return the address just past
	            the parsed text.

	    \return The integer value of encoded constant, if any is found,
	    else \c ULONG_MAX.

	    A constant at \e cp will be enclosed in single quotes.
	    On entry \e cp shall address the opening single quote.
	*/
	static unsigned long long
	read_encoded_char(int base, char const *cp, char const **pend);
};

#endif /* EOF*/
