#ifndef IO_H
#define IO_H

/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "platform.h"
#include "diagnostic.h"
#include <string>
#include <cassert>

/*!\file io.h
    This defines class \c io.
*/


/*! \brief Manages input and output file handling.

    Class \c io manages the input and output file-handling for each file
    that is processed and the I/O line-buffer.

*/
struct io {

	//! Nominal filename for the standard input stream.
	static char const * const _stdin_name_; // = "[stdin]"

	//! Get the start of the current line buffer.
	static char * line_start() {
		return const_cast<char *>(_wip_line_.c_str());
	}

	//! Get the end of the current line buffer.
	static char const * line_end() {
		return line_start() + _wip_line_.length();
	}

	//! Get a const reference to current line buffer as a string
	static std::string const & line_buffer() {
		return _wip_line_;
	}

	/*! \brief Get the offset of an address in the current line buffer.
	    \param  cp  An address in the line buffer.
	    \return The offset of \c cp in the line buffer.
	*/
	static ptrdiff_t read_offset(char const *cp) {
		return cp - line_start();
	}

	/*! \brief Read the name of a source file from \c stdin.

	    Filenames may contain spaces if quoted.

	    \return A pointer to the source filename in static storage,
	    if a valid filename is read; NULL if no filename is read.
	*/
	static std::string read_filename();

	/*! \brief Open a named file for reading or writing.

	    \param file The name of the file to be opened
	    \param mode	The read or write mode in which to open the file,
	    \c "r" or \c "w".
	    \return  Stream to opened file on success, else NULL.
	*/
	static FILE * open_file(std::string const & file, const char *mode);

	/*! \brief Open an input file and the appropriate output file.

	    \param		fname	The name of the input file to open.
	                An empty string, "", denotes the standard input
	*/
	static void open(std::string const & fname);

	/*! \brief Finalise the current pairing of source input and processed
	    output, if any.

	    \param		error	MSG_CLASS_NONE if no errors were raised in
	                processing the input file that is to be closed, else the
	                issue code  of the error that aborted processing.

	    If a file is associated with either of these streams, it is closed.
	    The output file replaces the input file, if \e error == MSG_CLASS_NONE
	    and the \c --replace option is in force, and is also backed up
	    beforehand, if the \c --backup  option is in force. If the function is
	    called before any input has been opened it is a no-op
	*/
	static void close(unsigned error);

	//! Close the current source file.
	static void close_input();

	/*! \brief Try to read a line of input from the current source file.
	    \return \e true iff a line is read.
	*/
	static bool get_line();

	//! Get a reference to the current source line.
	static std::string & wip_line() {
		return _wip_line_;
	}

	/*! \brief Read another line of input from the current source file.

	    Read another line of input from the current source file,
	    to extend the current line when a line-continuation is found or
	    when a newline is read within a C-comment.

	    \param	readpos	The current text pointer in the line buffer.
	    \return The restored text pointer, which will differ from \e
	     readpos if the line buffer has been reallocated.

	    The function assumes that more input is available. If not an
	    unexpected eof error is raised.

	*/
	static char * read_more(char const *readpos);

	/*! \brief Save the current line buffer position.

	    \param cp		The current line buffer position.

	    The position is saved as an offset into the line buffer
	    so that the correct position can be recovered if the
	    line buffer is moved.
	*/
	static void
	save_read_pos(char const *cp) {
		_saved_read_posn_ = cp - line_start();
	}


	/*! \brief Get the last saved line buffer position.

	    \returns cp	The line buffer position last saved by \e save_read_pos()

	    The returned position equates to the last position saved
	    even if the line buffer has moved in the interim.
	*/
	static char * saved_read_pos() {
		return line_start() + _saved_read_posn_;
	}

	//! Reset the input filename to \c NULL and line number to 0.
	static void top() {
		_line_num_ = 0;
		_in_filename_.resize(0);
	}

	/*! \brief Get the address of an offset into the input line buffer.
	    \param  readoff An offset into the line buffer.
	    \return The address of \c readoff.
	*/
	static char * read_pos(size_t readoff) {
		return line_start() + readoff;
	}

	//! Get the current source line number.
	static int line_num() {
		return _line_num_;
	}

	//! Get the name of the current source file.
	static std::string in_file_name() {
		return _in_filename_;
	}

	//! Get an unedited copy of the current input line.
	static std::string pristine_line();

	//! Get a pointer to the output stream.
	static std::ostream * output() {
		return _output_;
	}

	/*! \brief Set the directory in which to output a spin.
		\param	optarg	The commandline option specifing the spin directory
				name.
		The function stores the absolute path of the spin directory.
	*/
	static void set_spin_dir(char const *optarg);

	/*! \brief Set a path prefix (of input files) othat will be assumed to
	    match the name of the spin directory.
		\param	optarg	The commandline option specifing the spin prefix.

		The function stores the absolute path of the spin prefix.
	*/
	static void set_spin_prefix(char const *optarg);

	//! Get the name of operative spin directory.
	static std::string const & spin_dir() {
		return _spin_dir_;
	}

	//! Say whether there is any spin directory.
	static bool spin() {
		return !_spin_dir_.empty();
	}

private:

	/*!	\brief Read a source line from the current input stream into the
	    line buffer.
	    \return \e true iff a source line is read.
	*/
	static bool readon();

	/*!	\brief Replace the current input source file with the temporary output
	    file.

	    Replace the current input source file with the temporary output
	    file, when the \c --replace option is in force.

	    The input source file has either been deleted or renamed to a backup
	    when this function is called, depending on whether the \c --backup
	    option is in force.
	*/
	static void replace_infile();

	/*! \brief Generate a backup name for the current input source file.

	    Generate a backup name for the current input source file,
	    when the \c --replace and \c --backup options
	    are in force.

	    The backup filename is saved in static storage.

	    \param filename The name of the source file to be backed up.
	*/
	static void backup_infile(const char *filename);

	/*! \brief Backup the current input source file.

	    Backup the current input source file when the \c --backup option is
	    in force.

	    The current source file is renamed with a generated backup filename.
	*/
	static void backup_infile();

	/*! \brief Delete the current input source file.

	    Delete the current input source file, preparatory to replacing it
	    with the corresponding output file.
	*/
	static void delete_infile();


	//! Open the output file.
	static void open_outfile();

	/*! \brief Open an output stream for the current input file.
	    The output stream is \c cout unless input source
	    files are to be replaced with output. Otherwise the
	    output stream is opened on a temporary output file.
	*/
	static void open_output();

	/*! Try to read another line of input from the current source file.
	    \return \e true iff another line is read.
	*/
	static bool extend_line();

	/*! \brief Create an output file in the spin directory corresponding to
	    an input file.
	*/
	static void make_spinfile();

	/*! \brief Create a temporary file to which output will be written.

	    Create a temporary file to which output will be written
	    preparatory to replacing the input file with this output
	    when \c --replace is in effect.
	*/
	static void make_tempfile();

	/*! \brief Generate a filename for backing up an input file.

	    Generate a filename for backing up an input file
	    when \c --replace and \c --backup are in effect.
	    \param  filename The name of the input file to be backed up.
	*/
	static void make_backup_name(std::string const & filename);

	//! The current line number
	static int _line_num_;
	//! The name of the current source file
	static std::string _in_filename_;
	//! Unedited copy of the current input line for use in disgnostics
	static std::string _pristine_line_;
	//! The output stream
	static std::ostream * _output_;
	//! The input stream
	static std::istream * _input_;
	//! The input file
	static std::filebuf _infile_;
	//!  Current output filename, if needed
	static std::string _out_filename_;
	//! Backup filename, if needed
	static std::string _bak_filename_;
	//! Saved offset into line buffer
	static size_t _saved_read_posn_;
	//! The output file
	static std::filebuf _outfile_;
	//! The editable copy of the current input line
	static std::string _wip_line_;
	//! Name of directory in which to output a spin
	static std::string _spin_dir_;
	//! Path prefix assumed to match the spin directory
	static std::string _spin_prefix_;
};

#endif /* EOF*/
