#ifndef LINE_DESPATCH_H
#define LINE_DESPATCH_H

/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2002, 2003 Tony Finch <dot@dotat.at>.                   *
 *   All rights reserved.                                                  *
 *   Copyright (C) 1985, 1993 The Regents of the University of California. *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed by Mike Kinghan, imk@burroingroingjoing.com,              *
 *   derived from the code of Tony Finch                                   *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "platform.h"
#include <string>

/*!
    \file line_despatch.h
    This defines class \c line_despatch
*/

//! Enumeration of types of input lines
enum line_type {
    //! An hash-if that we can't resolve
    LT_IF,
    //! A true hash-if
    LT_TRUE,
    //! A false hash-if
    LT_FALSE,
    //! An hash-elif that we can't resolve
    LT_ELIF,
    //! A true hash-elif
    LT_ELTRUE,
    //! A false hash-elif
    LT_ELFALSE,
    //! An hash-else that we can't resolve
    LT_ELSE,
    //! An hash-endif
    LT_ENDIF,
    //! A non-directive line
    LT_PLAIN,
    //! End of file
    LT_EOF,
    //! A directive line of no more specific type than that is to be dropped
    LT_DIRECTIVE_DROP,
    //! A directive line of no more specific type that is to be kept
    LT_DIRECTIVE_KEEP,
    //! Sentinel
    LT_SENTINEL

};

/*! \brief Manages printing, discarding and replacement of lines.

    Class \c line_despatch class manages the printing of lines
    selected or generated for output and the dropping or
    replacing of lines not selected for output.
*/
struct line_despatch {

	//! Reinitialize the class static state.
	static void top() {
		_extension_lines_ = 0;
	}

	/*! \brief Increment the count of extension lines

	    Increment the count of extension lines i.e. physical lines that
	    logically continue the current source line.
	*/
	static void extend() {
		++_extension_lines_;
	}
	//!	Drop the line-buffer from output
	static void drop();

	//! Print the line-buffer to output
	static void print();

	/*! \brief Process the current input line and return its line type.
	    \return The \c line_type of the current input line.
	*/
	static line_type next();

	/*! \brief Substitute a diagnostic insert for the line in the line-buffer
	    and print it to output.

	    \param	replacement	The diagnostic insert to print.

	    Depending on the policy selected by the specified or default value of
	    the \c --conflict option, a \c \#define or \c \#undef directive read
		from input that contradicts one of the \c --define or \c --undefine 
		options may be replaced on output with a diagnostic comment or a
		diagnostic \c \#error directive.
	*/
	static void  substitute(std::string const & replacement);

	//! Get a reference to the count of suppressed lines.
	static unsigned & lines_suppressed() {
		return _lines_suppressed_;
	}

	//! Get a reference to the count of changed lines.
	static unsigned & lines_changed() {
		return _lines_changed_;
	}
private:

	//! Say whether the current line has been simplified.
	static bool line_simplified();

	//! Print an unmodified line buffer to output with no complications.
	static void printline_fast();

	/*! \brief Print a line buffer to output deleting chunks marked for
	    unconditional deletion.
	*/
	static void printline_cut();

	/*! \brief Print an output line buffer restoring deleted parentheses.

	    Print an output line buffer that may contain superfluous
	    parentheses marked for deletion, but no other parts
	    marked for deletion.

	    In this case we have not materially simplified the line and just restore
	    the parentheses before printing the line.
	*/
	static void printline_restore_all_paren();

	/*!	\brief Print an output line buffer containing chunks marked for
	    deletion.

	    We will delete all these chunks only if and only if some of them
	    represent truth-functional simplifications of the line.
	    Otherwise the only logical deletions will represent superfluous
		parentheses. We will then restore those parentheses before printing the
		line.
	*/
	static void printline_slow();

	/*!	\brief Print a line buffer to output.

	    The function selects a more specialised helper to which
	    it delegates.
	*/
	static void printline();

	//!	Print a line buffer to output commented out with the prefix "//coan".
	static void  printline_commented_out();


	/*! \brief Write a line to the output or drop it, according to command line
		options.

	    \param	keep	Is the line to be kept of dropped?

	    If \c keep is true then the line-buffer is output unless
	    the \c --complement option is in force.

	    If \c keep is false, or is true together with \c --complement,
	    the line buffer is dropped. If the \c --blank option is not in force
		then nothing is written to putput. Otherwise an empty line is written to
	    output.
	*/
	static void flushline(bool keep);

	//! Say whether the current line is to be output unsimplified.
	static bool line_unsimplified();

	//! Count of contiguous lines that are dropped together.
	static size_t _drop_run_length_;

	//! Number of linefeeds embedded in the current extended line.
	static unsigned _extension_lines_;
	//! Number of input lines suppressed.
	static unsigned _lines_suppressed_;
	//! Number of input lines changed
	static unsigned _lines_changed_;

};
#endif /* EOF*/
