#ifndef LINE_EDIT_H
#define LINE_EDIT_H

/***************************************************************************
 *   Copyright (C) 2004, 2006 Symbian Software Ltd.                        *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "platform.h"
#include <cstddef>

/*!
    \file line_edit.h
    This file defines class \c line_edit.
*/


/*! \brief Class \c line_edit class provides editing operations on the
    line-buffer.
*/
struct line_edit {

	/*! \brief Editing codes for deleted characters.

	    Character codes inserted into the
	    line-buffer in place of characters
	    provisionally marked for deletion.
	*/
	enum deletion_codes {
	    //! A character provisionally to be deleted.
	    DELETEABLE_CHAR = 1,
	    //! A left-parethesis provisionally marked for deletion.
	    DELETEABLE_LPAREN,
	    //! A right-parethesis provisionally marked for deletion.
	    DELETEABLE_RPAREN
	};

	/*! \brief Replace the keyword of the current directive.

	    Replace the keyword of the current directive with another
	    to rectify the syntax for dropped lines, then output
	    the edited directive.

	    \param	replacement		The keyword with which to replace the current
	                            keyword.
	*/
	static void keywordedit(const char *replacement);

	/*! \brief Logically delete redundant parentheses.

	    Logically delete a pair of redundant parentheses from a
	    directive in the line-buffer.

	    \param	lparen_off	    Line buffer offset of the '('
	    \param	rparen_off	    Line buffer offset of the ')'

	    Logical deletion is done by inserting \c DELETEABLE_LPAREN
	    in the line-buffer in place of the '(' and \c DELETEABLE_RPAREN
	    in place of the ')'.

	    If evaluation of the directive results in material simplification
	    (i.e. results in deletions other than deletions of redundant
		parentheses), then the logically deleted parentheses will be physically
		deleted when  evaluation is complete.

	    If evaluation of the directive results in no materal simplification
	    then the logically deleted parentheses will finally be restored
	    by restore_all_paren();

	*/
	static void delete_paren(size_t lparen_off, size_t rparen_off);

	/*! \brief Restore a prematurely deleted pair of parentheses in the
	        line buffer.

	    \param  lparen_off The offset of the left-hand parenthesis to be
			restored.
	    \param  rparen_off The offset of the right-hand parenthesis to be
			restored.

	    Recursive descent parsing obliges us to delete what appear to be
	    superfluous parentheses in the recursive context, but as recursion
	    unwinds it can transpire that deleted parentheses ought to have
	    been retained.
	*/
	static void restore_paren(size_t lparen_off, size_t rparen_off);

	/*! \brief Restore logically deleted parentheses to the directive.

	    Restore logically deleted parentheses to the directive in the
	    line buffer	when we cannot materially simplify it.

	    This is done on the principle that if we cannot materially simplify
	    the directive then wev will leave it alone.

	*/
	static void restore_all_paren();

	/*!	\brief Logically delete part of a directive in the line buffer.
	    \param	start	Offset to start of text to delete
	    \param	end		Offset terminating the text to delet.

	    The chunk to be deleted is only logically deleted by this function.
	    This is efficient because it avoids shuffling down terminal
	    segments of the buffer in the course of evaluation, when the effect of
	    such editing may be obliterated by subsequent deletions. Physical
		deletion is performed only once at the end of evaluation, by removing
		all logically  deleted portions of the line-buffer.

	    The function is only called to delete a truth-functionally redundant
	    operator and one of its operands.
	*/
	static void cut_text(size_t start, size_t end);

	/*! \brief Truncate the current line at the current text pointer.

	    \param	off	    Line buffer offset at which to truncate. This
	        becomes the new length of the line.
	*/
	static void truncate_line(size_t off);

	/*! \brief Say whether a character in the line buffer is deleteable.
	    \param  ch The character to be tested.
	    \return True if \c ch is an editing code inserted in the
	        line buffer in place of a character that can be deleted.
	*/
	static bool is_deleteable(char ch) {
		return ch >= DELETEABLE_CHAR && ch <= DELETEABLE_RPAREN;
	}

	//! Pointer to keyword of the current directive.
	static char*	_keyword_;
	//! True iff ops have been cut by simplification.
	static bool _ops_cut_;
	//! Count of the pairs of superflous parentheses deleted.
	static size_t _parens_deleted_;

private:

	/*! Replace text at the end of the line-buffer.

	    \param	where	The start of the terminal segment of the line-buffer
	                    to replace
	    \param	what	The start of the null-terminated string to replace
	                    the terminal segement at \c where.

	    The function does the editing work for \c keywordedit(). It ensures
	    that there is sufficient capacity in the line-buffer to contain the
	    replacement, extending the buffer if necessary.
	*/
	static void tail_edit(char *where, const char *what);

};

#endif /* EOF*/
