#ifndef OPTIONS_H
#define OPTIONS_H
/***************************************************************************
 *   Copyright (C) 2004 - 2006 Symbian Software Ltd.                       *
 *   All rights reserved.                                                  *
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/

#include "platform.h"
#include <string>
#include <vector>

/*! \file options.h
    This file provides the interface of the \e options class
*/

//! Symbolic constants denoting policies for discarding lines
enum discard_policy {
    DISCARD_DROP,	//!< Drop discarded lines
    DISCARD_BLANK,	//!< Blank discarded lines
    DISCARD_COMMENT //!< Comment discarded lines
};

/*! \brief Symbolic constants denoting policies for handling contradictions.

    Symbolic constants denoting policies for handling contradictions
    between commandline \c --define and \c --undef options and
    in-source \c \#define and \c \#undef directives.
*/
enum contradiction_policy {
    CONTRADICTION_DELETE,
    CONTRADICTION_COMMENT,
    CONTRADICTION_ERROR
};

//! Sequential symbolic constants for coan commands
enum command_code {
    CMD_HELP = 1,	//!< The help command
    CMD_VERSION,	//!< The version command
    CMD_SOURCE,	//!< The source command
    CMD_SYMBOLS,	//!< The symbols command
    CMD_INCLUDES,	//!< The includes command
    CMD_DEFS,	//!< The defs command
    CMD_PRAGMAS,	//!< The pragmas command
    CMD_ERRORS,	//!< The errors command
    CMD_DIRECTIVES,	//!< The directives command
    CMD_SPIN	//!< the spin command
};

//! Class \c options manages coan's commandline arguments
struct  options {

	//! Get the operative coan command code.
	static command_code get_command();

	//! Get our executable's full pathname, \c argv[0]
	static std::string const & exec_path() {
		return _exec_path_;
	}
	//! Get the program's name.
	static std::string const & prog_name() {
		return _prog_name_;
	}
	//! Get the file backup name suffix.
	static std::string const & backup_suffix() {
		return _backup_suffix_;
	}
	//! Have we finished parsing the options?
	static bool got_opts() {
		return _got_opts_;
	}
	//! Do we replace input files with output files?
	static bool replace() {
		return _replace_;
	}
	//! Do we report file and line numbers for listed items?
	static bool list_location() {
		return _list_locate_;
	}
	//! Do we report only the first occurrence of listed items?
	static bool list_only_once() {
		return _list_only_once_;
	}
	//! Do we list items only from operative directives?
	static bool list_only_active() {
		return _list_only_active_;
	}
	//! Do we list items only from inoperative directives?
	static bool list_only_inactive() {
		return _list_only_inactive_;
	}
	//! Do we list items only from inoperative directives?
	static bool list_symbols_in_ifs() {
		return _list_symbols_in_ifs_;
	}
	//! Do we list symbols in \c \#define directives?
	static bool list_symbols_in_defs() {
		return _list_symbols_in_defs_;
	}
	//! Do we list symbols in \c \#undef directives?
	static bool list_symbols_in_undefs() {
		return _list_symbols_in_undefs_;
	}
	//! Do we list symbols in \c \#include directives?
	static bool list_symbols_in_includes() {
		return _list_symbols_in_includes_;
	}
	//! Do we list system \c \#include directives?
	static bool list_system_includes() {
		return _list_system_includes_;
	}
	//! Do we list local \c \#include directives?
	static bool list_local_includes() {
		return _list_local_includes_;
	}
	//! Are we to output lines instead of dropping them and vice versa?
	static bool complement() {
		return _complement_;
	}
	/*! \brief Do we evaluate constants in truth-functional contexts or treat
	    them as unknowns.
	*/
	static bool eval_constants() {
		return _eval_consts_;
	}
	//! Do we report the resolved values of reported symbols?
	static bool resolve_symbols() {
		return _resolve_symbols_;
	}
	//! Get the policy for discarding lines.
	static discard_policy get_discard_policy() {
		return _discard_policy_;
	}
	//! Do we output \c \#line directives?
	static bool line_directives() {
		return _line_directives_;
	}
	//! Are we to omit parsing for C/C++ comments?
	static bool plaintext() {
		return _plaintext_;
	}
	//! Do we recurse into directories?
	static bool recurse() {
		return _recurse_;
	}
	//! Shall we continue to process input files after errors?
	static bool keep_going() {
		return	_keepgoing_;
	}
	//! Do we implicitly \c --undef all unconfigured symbols?
	static bool implicit() {
		return	_implicit_;
	}
	/*! Do we suppress transient symbol configurations for in-source
		\c \#define and \c \#undef directives? 
	*/
	static bool no_transients() {
		return	_no_transients_;
	}

	/*! \brief Parse the full and short names of the executable.

	    Element 0 of \e argv is parsed to store the
	    pathname and filename of our executable.

	    \param argv		The array of commandline arguments.
	*/
	static void parse_executable(char **argv);

	/*! \brief Parse the commandline.

	    \param argc		The number of commandline arguments.
	    \param	argv	Array of commandline arguments.
	*/
	static void parse(int argc, char *argv[]);

	//! Say whether a diagnostic reason code is gagged.
	static bool diagnostic_gagged(unsigned reason);

	/*! \brief Analyse the class global state after parsing the commandline.

	    Diagnose any errors and draw final conclusions.
	*/
	static void finish();

private:

	//! Symbolic constants for the options of all coan commands
	enum option_codes {
	    OPT_FILE = 'f', 		//!< The \c --file option
	    OPT_REPLACE = 'r', 		//!< The \c --replace option
	    OPT_BACKUP = 'b', 		//!< The \c --backup option
	    OPT_DEF = 'D', 			//!< The \c --define option
	    OPT_UNDEF = 'U', 		//!< The \c --undefine option
	    OPT_CONFLICT = 'x', 	//!< The \c --conflict option
	    OPT_GAG = 'g', 			//!< The \c --gag option
	    OPT_VERBOSE = 'V', 		//!< The \c --verbose option
	    OPT_COMPLEMENT = 'c', 	//!< The \c --complement option
	    OPT_EVALCONSTS = 'E', 	//!< The \c --constant option
	    OPT_DISCARD = 'k', 		//!< The \c --discard option
	    OPT_LINE = 2,			//!< The \c --line option
	    OPT_POD = 'P', 			//!< The \c --pod option
	    OPT_RECURSE = 'R', 		//!< The \c --recurse option
	    OPT_FILTER = 'F', 		//!< The \c --filter option
	    OPT_KEEPGOING = 'K',	//!< The \c --keepgoing option
	    OPT_IFS = 'i',			//!< The \c --ifs option
	    OPT_DEFS = 'd',			//!< The \c --defs option
	    OPT_UNDEFS = 'u',		//!< The \c --undefs option
	    OPT_INCLUDES = 3,       //!< The \c --includes option
	    OPT_LOCATE = 'L',		//!< The \c --locate option
	    OPT_ONCE = 'o',			//!< The \c --once option
	    OPT_SYSTEM = 's',		//!< The \c --system option
	    OPT_LOCAL = 'l',		//!< The \c --local option
	    OPT_ACTIVE = 'A',		//!< The \c --active option
	    OPT_INACTIVE = 'I',		//!< The \c --inactive option
	    OPT_EVALSYMS = 'e',		//!< The \c --evalsyms option
	    OPT_IMPLICIT = 'm',		//!< The \c --implicit option
	    OPT_DIR = 4,			//!< The \c --dir option
	    OPT_PREFIX = 'p',		//!< The \c --prefix option
	    OPT_NO_TRANSIENTS = 5	//!< The \c --no-transients option
	};

	/*! \brief Array of structures specifying the valid options for all coan
	    commands.
	*/
	static struct option long_options [];

	/*! \brief Array in which we will assemble the \c shortopts arguments for
	    \c getopts_long()

	    We allow 3 bytes per short option, i,e, the short option code
	    possibly followed by up to 2 colons.
	*/
	static char opts [];

	//! Array of structures specifying the valid coan commands.
	static struct cmd_option commands [];

	//! Excluded options for the \c source command.
	static int const source_cmd_exclusions [];

	//! Excluded options for the \c symbols command.
	static int const symbols_cmd_exclusions [];

	//! Excluded options for the \c includes command.
	static int const includes_cmd_exclusions[];

	//! Excluded options for the \c directives command.
	static int const directives_cmd_exclusions [];

	//! Excluded options for the \c defs command.
	static int const defs_cmd_exclusions [];

	//! Excluded options for the \c pragmas command.
	static int const pragmas_cmd_exclusions [];

	//! Excluded options for the \c errors command.
	static int const errors_cmd_exclusions[];

	//! Excluded options for the \c spin command.
	static int const spin_cmd_exclusions[];

	/*! \brief Array of exclusion lists for the coan commands,
	    indexed by command code.
	*/
	static struct exclusion_list const cmd_exclusion_lists [];

	/*! \brief Assemble the \c shortopts arguments for \c getopts_long().

	    The string is constructed by scanning the \c long_options
	    array to avoid the need for manually maintaining agreement
	    between that structure and the \c shortopts string..
	*/
	static void make_opts_list();

	/*!	\brief Configure diagnostic output.

	    Conigure diagnostic outputs in response to a \c --gag or \c --verbose
	    option.

	    \param	arg		Either "verbose" or an argument to the \c --gag option
	*/
	static void
	config_diagnostics(std::string const & arg);

	/*! \brief Configure the final state of diagnostic output filtering.

	    If no \c --gag or \c --verbose option has been seen,
	    default to \c --gag \c info \c --gag \c summary.
	*/
	static void finalise_diagnostics();

	/*! \brief Raise a usage error when an option is invalid for the
	    active command.

	    \param	cmd	 Pointer to details of the active coan command.
	    \param	bad_opt The code of the invalid option.
	*/
	static void
	error_invalid_opt(struct cmd_option const *cmd, int bad_opt);


	/*! \brief Add files to the input dataset.

	    \param path	File or directory to be included in the
	                input dataset.

	    If \c path is a file that satisfies any \c --filter option
	    it is added to the input dataset.

	    If \c path is a directory and \c --recurse is in force then
	    files recursively beneath it that satisfy any \c --filter
	    option are added to the input dataset.

	    If \c path is a directory and \c --recurse is not in force
	    then the directory is ignored.

	*/
	static void add_files(std::string const & path);

	/*!	\brief Parse commandline options from a file.

	    Parse commandline options specified with the \c --file
	    option.

	    \param	argsfile	Name of the file from which to read options
	    Options are separated by whitespace in the file.
	*/
	static void parse_file(std::string const & argsfile);

	/*!	\brief Parse the options to the active coan command.

	    \param	argc	The number of arguments to be parsed.
	    \param	argv	The array of arguments to be parsed.
	*/
	static void parse_command_args(int argc, char *argv[]);

	//! Say whether progress messages are suppressed.
	static bool progress_gagged();

	//! Pointer to the details of the operative coan command
	static struct cmd_option * _command_;
	//! \c argv[0]
	static std::string _exec_path_;
	//! Filename element of \c exec_path
	static std::string _prog_name_;
	//! Suffix for backup files
	static std::string _backup_suffix_;
	//! Are we finished parsing the commandline?
	static bool	_got_opts_;
	//! Do we replace input files with output files?
	static bool	_replace_;
	//! Do we report file and line numbers for listed items?
	static bool	_list_locate_;
	//! Do we report only the first occurrence of listed items?
	static bool	_list_only_once_;
	//! Do we list items only from kept lines?
	static bool	_list_only_active_;
	//! Do list items only from dropped lines?
	static bool	_list_only_inactive_;
	//! Do we list symbols in \c \#if/else/endif directives?
	static bool	_list_symbols_in_ifs_;
	//! Do we list symbols in \c \#define directives?
	static bool	_list_symbols_in_defs_;
	//! Do we list symbols in \c \#undef directives?
	static bool	_list_symbols_in_undefs_;
	//! Do we list symbols in \c \#include directives?
	static bool	_list_symbols_in_includes_;
	//! Do we list system \c \#include directives?
	static bool	_list_system_includes_;
	//! Do we list local \c \#include directives?
	static bool	_list_local_includes_;
	//! Are to output lines instead of dropping tem and vice versa?
	static bool	_complement_;
	/*! \brief Do we evaluate constants in truth-functional contexts or treat
	    them as unknowns?
	*/
	static bool	_eval_consts_;
	//! Do we report the resolved values of reported symbols
	static bool	_resolve_symbols_;
	//! Policy for discarding lines
	static discard_policy	_discard_policy_;
	//! Do we output \c \#line directives?
	static bool	_line_directives_;
	//! Are we to omit parsing for comments?
	static bool	_plaintext_;
	//! Recurse into directories?
	static bool	_recurse_;
	//! Continue to process input files after errors
	static bool	_keepgoing_;
	//! Do we implicitly \c --undef all unconfigured symbols?
	static bool	_implicit_;
	/*! Do we suppress transient symbol configurations for in-source
		\c \#define and \c \#undef directives? 
	*/
	static bool _no_transients_;
	//! Bitmask of diagnostic filters
	static int	_diagnostic_filter_;

	//! Array of options read from \c --file \c ARGFILE
	static std::vector<std::string > _argfile_argv_;
	//! Read whole \c ARGFILE into this storage
	static std::string 	_memfile_;
	//! The number of input files/dirs specified on the commandline.
	static int _cmd_line_files_;

	//! Are we parsing an argsfile ?
	static bool _parsing_file_;
};

#endif /* EOF */
