/***************************************************************************
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "parameter_list.h"
#include "chew.h"
#include "canonical_string.h"
#include "expression_parser.h"
#include "io.h"
#include "symbol.h"
#include "diagnostic.h"
#include <algorithm>

/*! \file parameter_list.cpp
    This file implements class \c parameter_list and class \c argument_list.
*/

using namespace std;

string parameter_list::str() const
{
	string list;
	if (_params.size()) {
		list.append(1,'(');
		vector<string>::const_iterator start = _params.begin();
		vector<string>::const_iterator const end = _params.end();
		for (	; start < end - 1; ++start) {
			list.append(*start);
			list.append(1,',');
		}
		list.append(*start);
		list.append(1,')');
	}
	return list;
}

void parameter_list::read(char const *cp, char const **pend)
{
	_params.empty();
	_well_formed = true;
	cp = chewer<continuation>(cp);
	if (*cp == '(') {
		cp = chewer<continuation>(++cp);
		while(*cp) {
			cp = chewer<discardable>(cp);
			string param = canonical<symbol>(cp,&cp);
			cp = chewer<discardable>(cp);
			if (cp[0] == '.') {
				cp = chewer<continuation>(++cp);
				if (*cp != '.') {
					_well_formed = false;
					break;
				}
				cp = chewer<continuation>(++cp);
				if (*cp != '.') {
					_well_formed = false;
					break;
				}
				_variadic = true;
				cp = chewer<discardable>(++cp);
			}
			if (*cp == ',' || *cp == ')') {
				if (param.length()) {
					_params.push_back(param);
				} else if (*cp == ',') {
					_well_formed = false;
					break;
				}
				if (*cp == ',') {
					if (_variadic) {
						_well_formed = false;
						break;
					}
					cp = chewer<continuation>(++cp);
				} else {
					break;
				}
			} else {
				_well_formed = false;
				break;
			}
		}
		if (*cp == ')') {
			++cp;
			if (_variadic) {
				_params.push_back("...");
			}
			if (!_params.size()) {
				_params.push_back(string());
			}
		}
	}
	if (pend) {
		*pend = cp;
	}
}

bool argument_list::append(string const & arg)
{
	if (arg.length()) {
		_params.push_back(arg);
		return true;
	}
	return false;
}


void argument_list::read(char const *cp, char const **pend)
{
	_params.empty();
	_well_formed = true;
	cp = chewer<continuation>(cp);
	if (*cp == '(') {
		cp = chewer<continuation>(++cp);
		size_t arg_off;
		char const *arg_start;
		bool resolved = false;
		integer resolved_val;
		while(*cp && *cp != ')') {
			arg_off = io::read_offset(cp);
			expression_parser ep(&cp,false);
			resolved = ep.result().resolved();
			if (resolved) {
				resolved_val = ep.result().value();
			}
			if (*cp == ',') {
				if (resolved) {
					append(diagnostic_base::citable_string(resolved_val));
				} else {
					arg_start = io::read_pos(arg_off);
					string arg = diagnostic_base::citable_string(arg_start,cp);
					if (!(_well_formed = append(arg))) {
						break;
					}
				}
				++cp;
			}
		}
		if (*cp == ')') {
			if (resolved) {
				append(diagnostic_base::citable_string(resolved_val));
			} else {
				arg_start = io::read_pos(arg_off);
				string arg = diagnostic_base::citable_string(arg_start,cp);
				bool more = append(arg);
				if (!more) {
					_well_formed = !_params.size();
				}
			}
			++cp;
		} else {
			_well_formed = false;
		}
	}
	if (pend) {
		*pend = cp;
	}
}

string
argument_list::subst(	parameter_list const & macro_params,
                        string const & definition) const
{
	string expansion = definition;
	size_t len = min(_params.size(),(size_t)macro_params.length());
	vector<string>::const_iterator param_iter = macro_params.begin();
	vector<string>::const_iterator param_end = param_iter + len;
	vector<string>::const_iterator arg_iter = _params.begin();
	for ( ; param_iter != param_end; ++param_iter,++arg_iter) {
		size_t offset = 0;
		for ( ; (offset = symbol::find_first_in(*param_iter,expansion,offset))
		      != string::npos;
		      offset += arg_iter->length()) {
			expansion.replace(offset,param_iter->length(),*arg_iter);
		}
	}
	return expansion;
}
