#ifndef PATH_H
#define PATH_H
/***************************************************************************
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "platform.h"
#include "cursor.h"
#include <utility>
#include <vector>
#include <string>
#include <stdexcept>

/*! \file path.h
    This file defines class \c path.
*/

/*! \brief Encapsulates a path.

    Class \c path controls a string representing an absolute or
    relative pathname and provides queries and operations on it.

    A \c path contains a \c \c cursor that can be used to maintain
    and modify an index into the elements of the path.

    \tparam Delim   The character that delimits elements of a path.
*/
template<char Delim>
struct path {
	path()
		: _first(0),_last(0) {}

	/*! \brief Explicitly construct a path object from a string.

	    \param  path  A string representation of a path.

	    The path's \c cursor is initialized to 0
	*/
	explicit path(std::string const & path)
		:   _path(path) {
		analyse();
	}

	//! Assign from a string.
	path & operator=(std::string const & str) {
		cut();
		_path = str;
		analyse();
		posn() = 0;
		return *this;
	}

	//! Equality operator.
	bool operator==(path const & other) const {
		return _path == other._path;
	}

	//! Inequality operator.
	bool operator!=(path const & other) const {
		return !(_path == other._path);
	}

	//! Less-than operator.
	bool operator<(path const & other) const {
		return _path < other._path;
	}

	/*! \brief Get the number of elements in the path

	    \return The number of elements in the the path.

	    The returned value is the number of \c Delim
	    characters in the path, even if there is a trailing
	    \c Delim.
	*/
	size_t elements() const {
		return _elements.size();
	}

	//! Get the path as a string.
	std::string const & str() const {
		return _path;
	}

	//! Get the length of the \c path.
	size_t length() const {
		return _path.length();
	}

	/*! \brief Return a path element by index.
	    \param which Index to the requested element
	    \return The indexed path element as a string.
	        If \e index is out of range an \e std::range_error
			exception is thrown.
	*/
	std::string element(size_t which) const {
		if (which >= elements()) {
			throw std::range_error("Out of range in path::element(size_t)");
		}
		element_pos const & element = _elements[which];
		return element.second == 0 ?
		       std::string(1,Delim) :
		       _path.substr(element.first,element.second);
	}

	//! Get the path element at the \c cursor.
	std::string cur_element() const {
		return element(posn());
	}

	//! The subscript operator is equivalent to \c element(which)
	std::string operator[](size_t which) const {
		return element(which);
	}

	/*! \brief Say whether the path consists of an initial subsequence
		of the elements of another.

		\param  other The \c path to be compared with \c *this.
		\return True if this path consists of an initial subsequence
		of the elements of \c other.
	*/
	bool is_prefix_of(path const & other) const {
		path common = common_prefix(*this,other);
		return *this == common;
	}

	/*! \brief Get a sub-sequence of the path's elements as a string.
	    \param  start   Index of the first element of the requested
	            sub-sequence.
	    \param  len Number of elements in the requested subsequence.

	    \return The requested sub-sequence as a string. If \e start is
	        out of range an empty string is returned. If there are fewer
	        than \e len elements at \e start then all remaining are
	        returned.
	*/
	std::string
	segment(size_t start = 0, size_t len = std::string::npos) const {
		element_pos section = get_section(start,len);
		return section.first != std::string::npos ?
		       _path.substr(section.first,section.second) : std::string();
	}

	/*! \brief Append a string to the path.

	    \param  str The string to append.

	    A trailing \c Delim character is appended if not already present,
		then \c str. \c str is not tested non-empty or for a leading \c Delim.
	*/
	void append(std::string const & str) {
		if (length() && _last != Delim) {
			_path += Delim;
		}
		_path +=  str;
		analyse();
	}

	/*! \brief Prepend a string to the path.

	    \param  str The string to prepend.

	    A leading \c Delim char is prepended if not already present,
		then \c str. \c str is not tested non-empty or for a trailing \c Delim.
	*/
	void prepend(std::string const & str) {
		if (_first != Delim) {
			_path.insert(0,1,Delim);
		}
		_path.insert(0,str);
		analyse();
	}

	/*! \brief Get the concatenation of the path with a string.

	    \param  str The string to be appended.
	    \return A new path that results from appending \e str to a copy
	    of \c *this with an intervening \c Delim.

		The new path inherits the \c cursor of \c *this
	*/
	path operator+(std::string const & str) const {
		path p(*this);
		p.append(str);
		return p;
	}

	/*! \brief Get the concatenation of the path with another.

	    \param  rhs The path to be appended.
	    \return A new path that results from appending \e rhs to a copy
	    of \c *this with amn intervening \c Delim.

	    The new path inherits the \c cursor of \c *this
	*/
	path operator+(path const & rhs) const {
		return *this + rhs.str();
	}

	//! \c operator+=(str) is equivalent to \c append(str)
	path & operator+=(std::string const & str) {
		append(str);
		return *this;
	}

	//! \c operator+=(path) is equivalent to \c append(path)
	path & operator+=(path const & rhs) {
		append(rhs.str());
		return *this;
	}

	/*! \brief Insert a string into the path.
	    \param  after The index of the element after which to insert.
	    \param  str The string to insert.

	    If \e after < elements() - 1, then \e str is inserted
	    following element \e after and delimited from the next
	    element by a \c Delim character. Otherwise the member
	    function is equivalent to \c append(str).
	*/
	void insert(size_t after,std::string const & str) {
		if (after < elements() - 1) {
			element_pos element = _elements[after];
			after = element.first + element.second + 1;
			_path.insert(after,str + Delim);
			analyse();
		} else {
			append(str);
		}
	}

	/*! \brief Remove a segment from the path.
	    \param  start   The index of the first element to remove.
	    \param  len     The number of elements to remove.

	    If \e start is out of range then nothing is done. If there
	    are fewer than \e len elements at \e start then all remaining
	    are removed.
	*/
	void cut(size_t start = 0, size_t len = std::string::npos) {
		element_pos section = get_section(start,len);
		if (section.first != std::string::npos) {
			if (section.first + section.second < _path.length()) {
				++section.second;
			} else if (section.first > 0) {
				--section.first;
			}
			_path.erase(section.first,section.second);
			analyse();
		}
	}

	/*! \brief Append a string to a \c path and \c cursor to the new last element.
	    \param  str The string to append.

	    The member function is like \c append but also updates
	    the \c cursor to the new final element.

	*/
	void push_back(std::string const & str) {
		append(str);
		posn() = elements() - 1;
	}

	/*! \brief Remove the last element of the path and decrement the \c cursor.

	    If the \c path is not empty, the final element is deleted and
	    the \c cursor is decremented to the bew final element.
	*/
	void pop_back() {
		size_t parts = elements();
		if (parts) {
			size_t cut = _elements[parts - 1].first;
			cut -= parts-- > 1;
			_path.erase(cut);
			analyse();
		}
		posn() = parts ? parts - 1 : 0;
	}

	/*! \brief Correct anomalies in the path.

	    All sequences of > 1 \c Delim are reduced to 1 and any trailing
	    \c Delim is removed.
	*/
	void rectify() {
		std::string match(2,Delim);
		for (size_t posn ; (posn = _path.find(match)) != std::string::npos; ) {
			_path.erase(posn + 1);
		}
		size_t len = _path.length();
		if (len > 1 && _path[len - 1] == Delim) {
			_path.resize(len - 1);
		}
		analyse();
	}

	///@{
	/*! \brief Get a [const] reference to the path's \c cursor.

	    \return A [const] reference to the \c cursor.

	    The \c cursor can be used to maintain and modify an
	    index into the path's elements.

	    No operation on a path except assignment modifies its
	    \c cursor, even if the operation leaves the \c cursor out of
	    range.
	*/
	cursor & posn() {
		return _cursor;
	}
	cursor const & posn() const {
		return _cursor;
	}
	///@}

	/*! \brief Get the common initial prefix of two paths

	    \param  lhs The first path to compare
	    \param  rhs The second path to compare
	    \return A \c path comprising the common initial elements of
	        \c lhs and \c rhs.
	*/
	static path common_prefix(path const & lhs, path const & rhs) {
		size_t count = std::min(lhs.elements(),rhs.elements());
		path p;
		for (size_t ndx = 0; ndx < count; ++ndx) {
			if (rhs[ndx] == lhs[ndx]) {
				p += rhs[ndx];
			}
		}
		return p;
	}

private:

	/*! \brief Type of an element or section locator in a \c path.

	    The type of an element or section locator in a \c path,
	    consisting of the index and length of the element.
	*/
	typedef std::pair<size_t,size_t> element_pos;

	//! Analyse the \c path to determine its composition.
	void analyse() {
		_elements.resize(0);
		size_t start = 0;
		size_t end;
		for (   ; (end = _path.find(Delim,start)) != std::string::npos;
		        start = end + 1) {
			_elements.push_back(element_pos(start,end - start));
		}
		if (start < _path.length()) {
			_elements.push_back(element_pos(start,_path.length() - start));
		}
		if (_path.length()) {
			_first = _path[0];
			_last = _path[_path.length() - 1];
		} else {
			_first = _last = 0;
		}
	}

	/*! \brief Get the location of a section of the \c path.
	    \param  start   The index of the first element of the required section.
	    \param  len     The number of elements in the required section. Defaults
	            to \c std::string::npos to request all elements from \c start.
	    \return An \c element_pos that delimits the required section.
	*/
	element_pos get_section(size_t start, size_t len = std::string::npos) const {
		if (len && start < elements()) {
			len = std::min(len,_path.length() - start);
			element_pos const & first = _elements[start];
			if (start + len <= elements()) {
				element_pos const & last = _elements[(start + len) - 1];
				len = (last.first + last.second) - first.first;
				return len ?
				       element_pos(first.first,len) : element_pos(first.first,1);
			}
			return element_pos(first.first,len);
		}
		return element_pos(std::string::npos,std::string::npos);
	}

	//! The string representation of the \c path.
	std::string _path;
	//! The structure of the path as a sequence of \c element_pos.
	std::vector<element_pos> _elements;
	//! The \c cursor of the path.
	cursor _cursor;
	//! The first character of the \c path.
	char _first;
	//! The last element of the \c path.
	char _last;
};



/*! \brief Get a path by concatenation of a string with a path.

    The \c cursor of the new path is initialized to 0
*/
template<char Delim>
path<Delim> operator+(std::string const & str, path<Delim> const & path)
{
	return path<Delim>(str) + path;
}

typedef path<PATH_DELIM> path_t;


#endif /* EOF*/
