#ifndef SYMBOL_H
#define SYMBOL_H
/***************************************************************************
 *   Copyright (C) 2007-2011 Mike Kinghan, imk@burroingroingjoing.com      *
 *   All rights reserved.                                                  *
 *                                                                         *
 *   Contributed originally by Mike Kinghan, imk@burroingroingjoing.com    *
 *                                                                         *
 *   Redistribution and use in source and binary forms, with or without    *
 *   modification, are permitted provided that the following conditions    *
 *   are met:                                                              *
 *                                                                         *
 *   Redistributions of source code must retain the above copyright        *
 *   notice, this list of conditions and the following disclaimer.         *
 *                                                                         *
 *   Redistributions in binary form must reproduce the above copyright     *
 *   notice, this list of conditions and the following disclaimer in the   *
 *   documentation and/or other materials provided with the distribution.  *
 *                                                                         *
 *   Neither the name of Symbian Software Ltd. nor the names of its        *
 *   contributors may be used to endorse or promote products derived from  *
 *   this software without specific prior written permission.              *
 *                                                                         *
 *   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 *   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 *   LIMITED TO, THE IMPLIED WARRANTIES OF  MERCHANTABILITY AND FITNESS    *
 *   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE        *
 *   COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,   *
 *   INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,  *
 *   BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS *
 *   OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED    *
 *   AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,*
 *   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF *
 *   THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH  *
 *   DAMAGE.                                                               *
 *                                                                         *
 **************************************************************************/
#include "platform.h"
#include "evaluation.h"
#include "line_despatch.h"
#include "parameter_list.h"
#include <memory>
#include <cstddef>
#include <cctype>
#include <map>

/*! \file symbol.h
    This file defines class \c symbol.
*/

//! Symbolic constants denoting the provenance of a symbol.
enum symbol_provenance {
    //! The symbol is not configured
    UNCONFIGURED,
    //! The symbol is globally configured by a commandline option
    CONFIGURED_GLOBAL,
    //! The symbol is transiently configured by an in-source directive
    CONFIGURED_TRANSIENT
};

//! Class \c symbol encapsulates a preprocessor symbol.
struct symbol {
	//! The default constructor creates the Null \c symbol */
	symbol()
		: _loc(_sym_tab_.begin()) {}

	//! Construct a \c symbol from a name.
	explicit symbol(std::string const & id)
		: _loc(_sym_tab_.find(id)) {
		if (_loc == _sym_tab_.end()) {
			_loc = insert(id,UNCONFIGURED);
		}
	}

	/*! \brief Construct a \c symbol given name and provenance.
	    \param  id The symbol identifier.
	    \param  source The \c provenance of the symbol.
	*/
	symbol(std::string const & id,
	       symbol_provenance source)
		: _loc(insert(id,source)) {}

	//! Equality operator
	bool operator==(symbol const & other) const {
		return _loc->first == other._loc->first &&
		       _loc->second == other._loc->second;
	}

	//! In equality operator
	bool operator!=(symbol const & other) const {
		return !operator==(other);
	}

	//! Get the symbol's name.
	std::string const & id() const {
		return _loc->first;
	}
	/*! \brief Get a pointer to the symbol's definition.
	    \return A pointer to the definition of the \c symbol, if any,
	    else \c NULL.
	*/
	std::string const * defn() const {
		return get_state()._defn;
	}

	//! Get the parameter list of the \c symbol.
	parameter_list const & macro_parameters() const {
		return get_state()._macro_params;
	}

	//! Say whether the symbol is defined.
	bool defined() const {
		return defn() != NULL;
	}

	//! Get the symbol's provenance.
	symbol_provenance provenance() const {
		return get_state()._provenance;
	}

	/*! \brief Get the source line-number at which this symbol was last defined
	    or undefined.
	*/
	unsigned source_line() const {
		return get_state()._line;
	}

	//! Say whether the symbol is configured.
	bool configured() const {
		return provenance() == CONFIGURED_GLOBAL ||
		       provenance() == CONFIGURED_TRANSIENT;
	}

	/*! \brief Say whether the symbol is global, i.e.
	    configured by the commandline options.
	 */
	bool global() const {
		return provenance() == CONFIGURED_GLOBAL;
	}

	/*! \brief Say whether the symbol has been reported with its current
	    definitional status.

	    \return True iff the symbol has been reported with its current
	    definitional status.

	    If the symbol's definitional status changes its reported
	    status becomes false.
	*/
	bool reported () const {
		return get_state()._reported;
	}

	//! Assign a definition to the symbol.
	void define(std::string const & defn = "") {
		get_state().define(defn);
	}

	//! Assign a macro parameter list to a symbol.
	void parameterize(parameter_list const & params) {
		get_state().parameterize(params);
		warn_variadic_unsupported();
	}

	//! Undefine the symbol.
	void undef() {
		get_state().undef();
	}

	/*! \brief Evaluate the symbol.

	    Compute the most-resolved textual form of a symbol and evaluate it.

	    \param  macro_args  The argument list with which the \c symbol is
	            invoked.

	    - If the symbol is neither defined nor undefined the symbol's name is
	        its most resolved form (MRF).
	    - If the symbol is undefined then the string "0" is the MRF.
	    - If the symbol if defined then its definition is recursively resolved
	    by substitution of the definitions of defined symbol references. In the
	    string that results from this exhaustive resolution, each maximally
	    long portion that is a soluble expression is evaluated and replaced
	    with its integer value. The result of this final replacement is the MRF.
	    - If the symbol's definition is circular then its definition is resolved
	    until the symbol's name appears in the resolved form, and this form is
	    the MRF.

	    \return An \c evaluation of the symbol's MRF.
	*/
	evaluation evaluate(argument_list const & macro_args) const {
		get_state().set_resolved_form(*this,macro_args);
		return get_state()._eval;
	}

	/*! \brief Get the symbol's most-resolved form (MRF)

	    Get the symbol's most-resolved form (MRF), as defined for \c evaluate().
	    \param macro_args  The argument list with which the \c symbol is
	            invoked.
	    \return A reference to the MRF of the \c symbol as a string.
	 */
	std::string const & resolve(argument_list const & macro_args) const {
		get_state().set_resolved_form(*this,macro_args);
		return *get_state()._resolved;
	}

	/*!	\brief Report an occurrence of the symbol.

	    Report an occurrence of the symbol
	    in accordance with the options of the \c symbols command.

	    \param  macro_args  The argument list with which the \c symbol is
	            invoked.

	    If this occurrence of the symbol is eligible for reporting it will
	    be listed on the standard output.

	    If \c --active is in force then symbols will be reported
	    only from active directives.

	    If \c --inactive is in force then symbols will be reported
	    only from inactive directives.

	    If \c --once is in force then a symbol will be reported only
	    once for each new definitional status.

	    If \c --locate is in force then the input filename and line number
	    is given with each reported occurrence.

	    If \c --ifs is in force then symbols will be listed from
	    \c if[[n][def]] and \c elif directives.

	    If \c --defs is in force then symbols will be listed from
	    \c define directives.

	    If \c --udefs is in force then symbols will be listed from
	    \c undef directives.
	*/
	void report(argument_list const & macro_args = argument_list()) const;

	//! \brief Record the symbol and its provenance in the symbol_table.
	void record(symbol_provenance source) {
		_loc = insert(id(),source);
	}

	/*! \brief Say whether this symbol is the Null symbol.

	    \return True iff this symbol is the Null symbol.

	    A symbol on which this method is called returns true is returned by
	    \c lookup(name) if there is no symbol named \c name in the
	    symbol table.
	*/
	bool null() const {
		return *this == symbol();
	}


	/*! \brief Analyse and handle a \c -D option for this symbol.

	    \param	cp	The option is of the form \c -DSYM[=VAL] or
	    \c --define \c SYM=[VAL]. \c cp shall point just past \c SYM.

	    The option is analysed for well-formedness and for consistency and
	    non-redundancy in the current configuration. Errors are diagnosed.
	    If there are none the global configuration is updated if required.
	 */
	void digest_global_define(char const *cp);

	/*! \brief Analyse and handle an in-source \c \#define directive for this
	    symbol.

	    \param   macro_params The macro parameter list of the symbol.
	    \param	definition	The proposed definition of the symbol.
	    \return A \c line_type resulting determined by the analysis, denoting
	    the action to be taken on the directive.

	    The proposed defintion is analysed for consistency and
	    non-redundancy in the current configuration. Errors are diagnosed.
	    If there are none the transient configuration is updated if required.
	 */
	line_type
	digest_transient_define(parameter_list const & macro_params,
	                        std::string const & definition);


	/*! \brief Analyse and handle a \c -U option for this symbol.

	    \param	cp	The option if of the form \c -USYM. \c cp shall
	    point just past \c SYM.

	    The option is analysed for well-formedness and for consistency and
	    non-redundancy in the current configuration. Errors are diagnosed.
	    If there are none the global configuration is updated if required.
	 */
	void digest_global_undef(char const *cp);

	/*! \brief Analyse and handle an in-source \c \##undef directive for this
	    symbol.

	    \return A \c line_type denoting the action to be taken with the
	        directive.
	    The directive is analysed for well-formedness and for consistency
	    non-redundancy in the current configuration. Errors that are diagnosed.
	    If there are none the transient configuration is updated if required.
	 */
	line_type digest_transient_undef();


	/*! \brief Expand an invocation of this symbol.
	    \param macro_args   The argument list with which the symbol is invoked.
	    \return The expansion of the symbol invocation as a string.
	*/
	std::string macro_expand(argument_list const & macro_args) const;


	/*! \brief Read an identifier from a text pointer.

	    \param cp   Address of the identifier.
	    \param pend  If not NULL, receives on return the address just past
	        the identifier.

	    \return The parsed identifier.
	*/
	static std::string read(char const *cp, char const **pend = NULL);

	/*! \brief Lookup a symbol by name in the symbol table.

	    \param  id    The name of the symbol sought.
	    \return A symbol. If no symbol named \c id is found then
	        the Null symbol is returned.
	*/
	static symbol lookup(std::string const & id) {
		symbol_table::iterator result = _sym_tab_.find(id);
		return result == _sym_tab_.end() ?
		       symbol() : symbol(result);
	}

	//! Say whether a character can be the first of a symbol name.
	static bool is_start_char(char ch) {
		return (unsigned)ch < 255 && (isalpha(ch) || ch == '_');
	}

	//! Say whether a character can occur in symbol name.

	static bool is_valid_char(char ch) {
		return (unsigned)ch < 255 && (isalnum(ch) || ch == '_');
	}

	//! Delete all transient symbols from the symbol table
	static void purge_transients();

	//! Get the number of symbols with a given provenance.
	static size_t count(symbol_provenance source);

	/*! \brief Find the first occurrence of a symbol within a terminal segment
		of a string.

	    \param name The name of the symbol to be searched for.
	    \param str  The string in which to search for \c name.
	    \param start The offset in \c str at which to start searching.
	    \return The offset in \c str of the first occurrence of \c name at or
	        after \c start, if any, else \c string::npos.

	*/
	static size_t
	find_first_in(std::string name, std::string str, size_t start);

	/*! \brief Scan for a symbol name from a text pointer.
	    \param  cp  The address at which to start scanning.
	    \return The address just passed any scanned symbol name.
	*/
	static char const * scan_name(char const * cp);

private:

	//! Encapsulates the symbol's state */
	struct state {
		/*! \brief Explicitly construct a symbol \c state given a \c provenance.
		    \param source   The \c provenance of the symbol to construct.
		*/
		explicit state(symbol_provenance source)
			:   _defn(NULL),
			    _provenance(source),
			    _resolved(NULL),
			    _reported(false),
			    _line(0) {}

		//! Copy constructor.
		state(state const & src)
			:   _defn(src._defn ? new std::string(*src._defn) : NULL),
			    _macro_params(src._macro_params),
			    _provenance(src._provenance),
			    _resolved(src._resolved ?
					new std::string(*src._resolved) : NULL),
			    _reported(src._reported),
			    _line(src._line) {}

		//! Swap operation.
		void swap(state & other) {
			std::swap(_defn,other._defn);
			std::swap(_macro_params,other._macro_params);
			std::swap(_provenance,other._provenance);
			std::swap(_resolved,other._resolved);
			std::swap(_reported,other._reported);
			std::swap(_line,other._line);
		}

		//! Assignment.
		state & operator=(state other) {
			swap(other);
			return *this;
		}

		//! Destructor
		~state() {
			delete _defn;
			delete _resolved;
		}

		//! Equality operator
		bool operator==(state const & other) const {
			return ((!_defn && !other._defn) || (*_defn == *other._defn)) &&
			       _macro_params == other._macro_params &&
			       _provenance == other._provenance;
		}

		//! Inequality operator
		bool operator!=(state const & other) const {
			return !operator==(other);
		}


		//! Set a definition in a symbol's \c state.
		void define(std::string const & defn);

		//! Set a macro parameter list in a symbol's \c state.
		void parameterize(parameter_list const & params) {
			_macro_params = params;
		}

		//! Remove any symbol definition from a symbol's \c state.
		void undef();

		/*! \brief Set the most resolved form (MRF) in a symbol's \c state.
		    \param sym The symbol that owns this \c state.
		    \param  macro_args  The argument list with which \c sym is
		        invoked.
		*/
		void
		set_resolved_form(	symbol const & sym,
		                    argument_list const & macro_args) const;

		//! Pointer to the symbol's definition, if any
		std::string * _defn;
		//! List of the symbol's macro arguments, if any
		parameter_list _macro_params;
		//! The provenance of the symbol
		symbol_provenance _provenance;
		//! Pointer to the symbol's most resolved form, if computed
		mutable std::string * _resolved;
		//! Has this symbol been reported?
		mutable bool _reported;
		//! Evaluation of the symbol's most-resolved form
		mutable evaluation _eval;
		//! Source line-number where symbol last defined or undefined, or 0.
		mutable unsigned _line;

		//! Are symbol resolutions out of date?
		static bool _out_of_date_;
	};

	//! Type of the symbol table.
	typedef std::map<std::string,state> symbol_table;
	//! Type of a symbol table entry.
	typedef symbol_table::value_type table_entry;
#if !CPP11_HAVE_AUTO_DECL
	typedef std::pair<symbol_table::iterator,bool> sym_table_find_result;
#endif
 
	//! Privately construct a \c symbol from a symbol table iterator.
	explicit symbol(symbol_table::iterator iter)
		: _loc(iter) {}

	///@{
	//! \brief Get a [const] reference to the symbol's state.
	state & get_state() {
		return _loc->second;
	}
	state const & get_state() const {
		return _loc->second;
	}
	///@}

	/*! \brief Insert the symbol into the symbol table with a specified
	    provenance.
	    \param  id The name of the symbol to insert.
	    \param  source The \c provenance of the symbol.
	    \return An iterator to the inserted symbol.
	*/
	symbol_table::iterator
	insert(std::string const & id, symbol_provenance source) {
#if CPP11_HAVE_AUTO_DECL
		auto result = _sym_tab_.insert(table_entry(id,state(source)));
#else
		sym_table_find_result result =
			_sym_tab_.insert(table_entry(id,state(source)));
#endif
		result.first->second._provenance = source;
		return result.first;
	}

	/*! \brief Search a terminal portion of a string to find the first sytactic
	    occurrence of the symbol's name.

	    \param  str The string to search.
	    \param  start   The offset in \e str at which to start searching.
	    \return The offset in \e str of the occurrence located, if any,
	        else \c string::npos.
	*/
	size_t find_first_in(std::string const & str, size_t start) const {
		return find_first_in(id(),str,start);
	}

	/*! \brief Search a terminal portion of a string for any configured symbol
	    name.

	    Search a terminal portion of a string to find the first syntactic
	    occurrence of the name of any configured symbol.

	    \param  str  The string to search.
	    \param  start   On entry, the offset in \e str
	        at which to start searching. On return receives the offset
	        of the occurrence located, if any, and otherwise is unchanged.

	    \return A symbol sym s.t. that \c sym.null() is true if no
	    configured symbol's name is found.
	*/
	static symbol find_any_in(std::string const & str, size_t & start);
	
	//! Apologize that variadic macros are unsupported 
	void warn_variadic_unsupported();

	//! Iterator to the symbol in the symbol table.
	symbol_table::iterator _loc;

	//! The symbol table.
	static symbol_table _sym_tab_;
	//! The table entry of the null symbol
	static table_entry _null_symbol_entry_;

};

/*@}*/

#endif /* EOF*/
